/************************************************************************
 * This library is free software; you can redistribute it and/or	*
 * modify it under the terms of the GNU Library General Public		*
 * License as published by the Free Software Foundation; either		*
 * version 2 of the License, or (at your option) any later version.	*
 * 									*
 * This library is distributed in the hope that it will be useful,	*
 * but WITHOUT ANY WARRANTY; without even the implied warranty of	*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU	*
 * Library General Public License for more details.			*
 * 									*
 * You should have received a copy of the GNU Library General Public	*
 * License along with this library; if not, write to the		*
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,		*
 * Boston, MA  02111-1307, USA.						*
 * 									*
 * The GNU Library General Public License file is included in		*
 * Online-Documentation							*
 *									*
 * GNU's Homepage - http://www.gnu.ai.mit.edu/home.html			*
 *									*
 * Questions & comments referring to *FLIP++: please contact us:	*
 *									*
 * Snail-Mail:								*
 *	Dr. Wolfgang SLANY						*
 *	Database and Expert System Group (E184-2)     			*
 *	Institute of Information Systems               			*
 *	Vienna University of Technology					*
 *	Paniglgasse 16, A-1040 Vienna, AUSTRIA				*
 * Telephone:								*
 *	+43-1-58801-6123     (area-code 1 maybe changed to 201 in 1998) *
 * FAX:									*
 *	+43-1-5055304	     (area-code 1 maybe changed to 201 in 1998) *
 * eMail:								*
 *	sflip@dbai.tuwien.ac.at						*
 * WWW:									*
 *	http://www.dbai.tuwien.ac.at/proj/StarFLIP/			*
 *                                                                      *
 * $Id: expint.h,v 1.4 1997/06/24 17:29:57 reichel Exp $ *
 ***********************************************************************/


// -----------------------------------------------------------------------
//	What   : an expression interpreter
//	Project: DynaFLIP++
//	Author : Ulrich Santa
//	Start  : 13.12.1995
//	Last   : 28.06.1996
//      File   : expint.h
// -----------------------------------------------------------------------
// Operators:		+,-,*,/
// Compares:		=,<,>,!
// Conditions:		?(expr Compares expr) (expr) (expr)
// Functions:  	SIN(expr),COS(),TAN(),SQR(),SRT(),ABS(),LOG(),MIN(),MAX()
// Aggregators:   SUM{expr[,expr]},MUL{expr[,expr]}
// 					SUM{Varname[i]}/n -> Varname and n will be substituted
// Signs have to be put into brackets (-1), (-(SIN(2))
// -----------------------------------------------------------------------
#ifndef EXPINT
#define EXPINT

#include <stdlib.h>
#include <stdio.h>
#include <fstream.h>
#include <string.h>
#include <math.h>

#define MAX_FORMULA_LEN 128
#define MAX_NAME_LEN 80
#define NODE 1
#define LEAF_VALUE 2
#define LEAF_VARIABLE 3
#define TRUE 1
#define FALSE 0
#define SMART 1
#define ALWAYS 0

typedef char string[MAX_FORMULA_LEN];

struct Tindex
 {
  char name[2];
  unsigned int start;
  unsigned int end;
  unsigned int real_found;
 };


/*-----------------------------------------------------------------------*/
/* Linear list for variable handling                                     */
/*-----------------------------------------------------------------------*/
class TVarlist
{
private:
  char name[MAX_NAME_LEN];
  float value;
  int computed;
  char formula[MAX_FORMULA_LEN];
  TVarlist *next;
  unsigned int entries;
public:
	TVarlist(void)
		{ set_name("");set_value(0);set_comp(FALSE);set_formula("");next = NULL; entries = 0;};
	TVarlist (char *n, char *f, float v)
		{ set_name(n); set_value(v); set_formula(f); next = NULL; entries = 0;};
	~TVarlist(void) {};
	void set_name(char *n) {strcpy(name,n);};
	void set_formula(char *f) {strcpy(formula,f);};
	void set_value(float v) {value = v;};
	void set_comp(int c) { computed = c;};
	int get_comp(void) {return computed;};
	void set_comp(char *,int);
	int get_comp(char*);
	unsigned int get_entries(void) {return entries;};
	char *get_name(void) {return name;};
	char *get_formula(void) {return formula;};
	float get_value(void) {return value;};
	void insert(char *, char *,float );
	char *get_formula(char *);
	void set_formula(char*,char*);
	float get_value(char *);
	void set_value(char*,float);
	void remove(char *);
	char *check_expr(char *);
	char *substitute(char *);
	char *real_name(char *);
	void substitute_all(void);
	TVarlist *clean_up(void);
	void show(void);
	void show(char *);
	int in_list(char *);
        TVarlist *get_var(char*);
	void read(ifstream &);
	void write(ofstream &);
	void read(char*);
	void write(char*);
	TVarlist* get_next(void) {return next;};
};


/*-----------------------------------------------------------------------*/
/* Parse-tree-object for mathematical expressions                        */
/*-----------------------------------------------------------------------*/
class TExptree
{
private:
 int  type;								// 1 .. Node   2 .. Leaf
 int sign;								// -1 or +1 as multiplicator
 char oper[10];						// operator such as {+,-,*,/}
 char func[10];                  // function such as {SIN,COS,SQR,SUM,...}
 float value;							// Value
 char variable[MAX_NAME_LEN];		// Name of Variable
 TExptree *left;
 TExptree *right;
public:
 TExptree(void);
 TExptree(int t,char *op,char *f,char *v,float val,TExptree *l,TExptree *r)
	 { set_type(t); set_sign(1); set_oper(op); set_function(f);
		set_variable(v); set_value(val); set_left(l); set_right(r);};
 ~TExptree() {};
 void set_type(int t) { type = t;};
 void set_oper(char *op) { strcpy(oper,op);};
 void set_sign(int s) { sign = s;};
 void set_value(float val) { value = val;};
 void set_function(char *f) {strcpy(func,f);};
 void set_left(TExptree *l) { left = l;};
 void set_right(TExptree *r) { right = r;};
 void set_variable(char *v) {strcpy(variable,v);};
 char *get_variable(void) { return variable;};
 int get_type(void) { return type;};
 char *get_oper(void) {return oper;};
 int get_sign(void) {return sign;};
 float get_value(void) { return value;};
 float string_to_float(char*);
 char *get_function(void) { return func;};
 TExptree *get_left(void) { return left;};
 TExptree *get_right(void) { return right;};
 void insert_expr(char *, int );
 char *get_next(char*);
 void show_tree(int, int);
 TExptree *clean_up(TExptree*);
 void show(int what=LEAF_VALUE) { show_tree(0,what);};
 float calculate(TVarlist *,int=ALWAYS);
 void avoid_recursion(char*,TVarlist*);
};

#endif

