/************************************************************************
 * This library is free software; you can redistribute it and/or	*
 * modify it under the terms of the GNU Library General Public		*
 * License as published by the Free Software Foundation; either		*
 * version 2 of the License, or (at your option) any later version.	*
 * 									*
 * This library is distributed in the hope that it will be useful,	*
 * but WITHOUT ANY WARRANTY; without even the implied warranty of	*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU	*
 * Library General Public License for more details.			*
 * 									*
 * You should have received a copy of the GNU Library General Public	*
 * License along with this library; if not, write to the		*
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,		*
 * Boston, MA  02111-1307, USA.						*
 * 									*
 * The GNU Library General Public License file is included in		*
 * Online-Documentation							*
 *									*
 * GNU's Homepage - http://www.gnu.ai.mit.edu/home.html			*
 *									*
 * Questions & comments referring to *FLIP++: please contact us:	*
 *									*
 * Snail-Mail:								*
 *	Dr. Wolfgang SLANY						*
 *	Database and Expert System Group (E184-2)     			*
 *	Institute of Information Systems               			*
 *	Vienna University of Technology					*
 *	Paniglgasse 16, A-1040 Vienna, AUSTRIA				*
 * Telephone:								*
 *	+43-1-58801-6123     (area-code 1 maybe changed to 201 in 1998) *
 * FAX:									*
 *	+43-1-5055304	     (area-code 1 maybe changed to 201 in 1998) *
 * eMail:								*
 *	sflip@dbai.tuwien.ac.at						*
 * WWW:									*
 *	http://www.dbai.tuwien.ac.at/proj/StarFLIP/			*
 *                                                                      *
 * $Id: flipcvar.cpp,v 1.4 1997/06/24 17:30:00 reichel Exp $ *
 ***********************************************************************/


//
// $Id: flipcvar.cpp,v 1.4 1997/06/24 17:30:00 reichel Exp $
// 
//
//     ------------------------------------
//	F L I P C V A R . C C  ( . C P P )
//     ------------------------------------
//

#include "flipcvar.h"
 
//
//
//	* * * * * * * * * *  F L i s t  * * *
//
//

// * * destructor:

 FList::~FList(void) {
	while (first != NULL) {
		FloatValue *e=first;
		first = e->get_next();
		delete e; }; };

// * * complex methods:

void FList::add(float new_v) {
	FloatValue *element=new FloatValue;
	element->set_value(new_v);
	if (first == NULL ) { first=element;           last=element; }
	else 		    { last->set_next(element); last=element; };
	number++; };

FloatValue *FList::find(int index) {
	FloatValue *element=first;
	if (index < number) {
		for (int i=0; i<index; i++) element=element->get_next();
		return element; }
	else return NULL; };

float FList::find_v(int index) {
	FloatValue *v=find(index);
	if (v!=NULL) return v->get_value();
	else	     return -1; };

int FList::remove(int index) {
	if (index >= number) {
		cerr << "no_such_FloatValue" << endl; 
		return -1; };
	FloatValue *element;
	FloatValue *to_delete;
	if (index==0) { to_delete=first;
			  first=first->get_next(); }
	else { element=find(index-1);
	       to_delete=element->get_next();
	       element->set_next(element->get_next()->get_next()); };
	delete to_delete;
	number--;
	return 1; };

void FList::remove_all(void) {
	while (first != NULL) {
		FloatValue *n=first;
		first = n->get_next();
		delete n; };
	first=NULL;
	last=NULL;
	number=0; };

float FList::mean(void) {
	FloatValue *element=first;
	float sum=0;
	while (element!=NULL) {
		sum += element->get_value();
		element=element->get_next(); };
	return sum/number; };

//
//
//	* * * * * * * * * *   P o i n t   * * *
//
//

// * * complex methods:

void Point::show(void) {
	cout << x_value << '/' << y_value << space; };

int Point::write(ofstream *output) {
	*output << x_value << space << y_value << space;
	return 1; };

int Point::read(ifstream *input) {
	*input >> x_value >> y_value;
	return 1; };

//
//
//	* * * * * * * * * *  P L i s t  * * *
//
//

// * * destructor:

 PList::~PList(void) {
	while (first != NULL) {
		Point *n=first;
		first = n->get_next();
		delete n; }; };

// * * complex methods:

void PList::show(void) {
	Point *element=first;
	while (element != NULL) {
		element->show();
		element=element->next; }; };


int PList::write(ofstream *output) {
	Point *element=first;
	*output << "/* #: " << number << " */ ";
	while (element != NULL) {
		element->write(output);
		element=element->next; };
	return 1; };

int PList::read(ifstream *input) {
	char instr[10];
	int n;
	*input >> instr; //  /*
	*input >> instr; //  #:
	*input >> n;
	*input >> instr; //  */
	for (int i=0; i<n; i++) {
		add();
		last->read(input); };
	return 1; };

void PList::add(float new_x, float new_y) {
	Point *element=new Point;
	element->x_value=new_x;
	element->y_value=new_y;
	if (first == NULL ) { first=element;      last=element; }
	else 		    { last->next=element; last=element; };
	number++; };

Point *PList::find(int index) {
	Point *element=first;
	if (index < number) {
		for (int i=0; i<index; i++) element=element->next;
		return element; }
	else return NULL; };

int PList::find(float x,float y) {
	Point *element=first;
	int index=0;
	while (element->get_next()!=NULL && element->get_x()!=x) {
		element=element->get_next();
		index++; };
	float ex=element->get_x();
	float ey=element->get_y();
	if (ex==x && ey==y)
		return index;
	else    return -1; };

int PList::insert(float new_x, float new_y) {
	Point *element=first;
	Point *before_element=first;
	if (number<1) {
		add(new_x,new_y);
		return 1; };
	if (find(new_x,new_y) >= 0) {
		return 0; };
	Point *new_element=new Point;
	new_element->x_value=new_x;
	new_element->y_value=new_y;
	number++;
	while (element->next != NULL && element->next->x_value <= new_x) {
		before_element=element;
		element=element->next; };
	// now element->x_value <= new x or element == last
	float element_y=element->get_y();
	float element_x=element->get_x();
	if ( element_x == new_x ) {
		if ( element == first )
			if ( element_y > new_y ) {
				// insert before first
				new_element->next=element;
				first=new_element;
				return 2; }
	    		else {
				// insert after element
				new_element->next=element->next;
				element->next=new_element;
				if ( element == last ) last=new_element;
				return 3; };
	  	if ( element == last )
	    		if ( element_y < new_y ) {
				// insert before element
				new_element->next=element;
				if (before_element == first)
					first = new_element;
				else
					before_element->next=new_element;
					return 4; }
	    		else {
				// insert after element
				new_element->next=element->next;
				element->next=new_element;
				if ( element == last ) last=new_element;
				return 5; }; };
	// insert after element
	new_element->next=element->next;
	element->next=new_element;
	if ( element == last ) last=new_element;

	float max_y  = get_max_y();
	if ( element != last && element != first) {
		float prev_y = element->get_y();
		float next_y = element->next->get_y();
		float low_y = min(next_y,prev_y);
		if (new_y < low_y) new_element->set_y(low_y);
		float high_y = max(next_y,prev_y);
		if (high_y < max_y && new_y > high_y) new_element->set_y(high_y); 
		}; 

	return 6; };


void PList::remove(int index) {
	Point *element;
	Point *to_delete;
	if (index < number) {
		if (index==0) { to_delete=first;
				first=first->next; }
		else { element=find(index-1);
		       to_delete=element->next;
		       element->next=element->next->next; };
		number--;
		delete to_delete; }; };

void PList::remove_all(void) {
	while (first != NULL) {
		Point *n=first;
		first = n->get_next();
		delete n; };
	first=NULL;
	last=NULL;
	number=0; };

void PList::copy(PList *p_list) {
	remove_all();
	Point *element=p_list->get_first();
	while (element!=NULL) {
		add(element->get_x(),element->get_y());
		element=element->get_next(); }; };

float PList::get_max_y(void) {
	float max=0.0;
	int   i=0;
	while (find(i)!=NULL) {
		float v=find_y(i);
		if (v>max) max=v;
		i++; };
	return max; };

float PList::get_first_x(float y) {
	Point *element=first;
	while (element!=NULL) { 
		if (element->get_y() == y) return element->get_x();
		element=element->get_next(); }; 
	cerr << "y-value not found in PList::get_first_x(float y)" << endl;
	return -1; };

void PList::get_x(FList *xl, float y) {
	Point *element=first;
	while (element!=NULL) {
		if (element->get_y() == y) xl->add(element->get_x());
		element=element->get_next(); }; };
	
	
void PList::clear_horizontal_line(void) {
	// remove all points between two endpoints in horizontal line
	if (number > 2) {
		Point* before=first;
		Point* element=before->get_next();
		Point* after=element->get_next();
		while (after!=NULL) {
			if (before->get_y()==element->get_y() &&
			    element->get_y()==after->get_y())
				remove(find(element));
			before =before->get_next();
			element=before->get_next();
			if (element != NULL) after = element->get_next();
			    else after = NULL;
			 }; }; };

void PList::clear_vertical_line(void) {
	// remove all points between two endpoints in vertical line
	if (number > 2) {
		Point* before=first;
		Point* element=before->get_next();
		Point* after=element->get_next();
		while (after!=NULL) {
			if (before->get_y()==element->get_x() &&
			    element->get_y()==after->get_x())
				remove(find(element));
			before =before->get_next();
			element=before->get_next();
			if (element != NULL) after = element->get_next();
			    else after = NULL;
			 }; }; };

//
//
//	* * * * * * * * * *   F u z z y S e t B a s e   * * *
//
//

// * * constructors:

FuzzySetBase::FuzzySetBase(char *init_linguistic_term,
		 float init_lower_support_end, float init_upper_support_end,
		 colourtype init_fs_colour) {
	strcpy(linguistic_term, init_linguistic_term);
	lower_support_end = init_lower_support_end;
	upper_support_end = init_upper_support_end;
	fs_colour.set_colour(init_fs_colour); };

FuzzySetBase::FuzzySetBase() {
	strcpy(linguistic_term, "no_linguistic_term");
	lower_support_end = 0.0;
	upper_support_end = 0.0;
	fs_colour.set_colour(white); };

// * * complex methods:

void FuzzySetBase::show(void) {
	cout << " linguistic_term: " << linguistic_term << endl
	     << "lower_support_end: " << lower_support_end << tab
	     << "upper_support_end: " << upper_support_end << endl
	     << "fs_colour: ";
	fs_colour.show();
	cout << tab; };

int FuzzySetBase::write(ofstream *output) {
	*output << linguistic_term << space << lower_support_end << space
		<< upper_support_end << space;
	fs_colour.write(output); *output << space;
	return 1; };

int FuzzySetBase::read(ifstream *input) {
	*input >> linguistic_term >> lower_support_end
	       >> upper_support_end;
	fs_colour.read(input);
	return 1; };

void FuzzySetBase::copy(FuzzySetBase *fsb) {
	strcpy(linguistic_term,fsb->linguistic_term);
	lower_support_end = fsb->lower_support_end;
	upper_support_end = fsb->upper_support_end;
	fs_colour.set_colour(*(fsb->fs_colour.get_colour())); };

//
//
//	* * * * * * * * * *   P l i m F u z z y S e t   * * *
//
//
// * * constructors:

PlimFuzzySet::PlimFuzzySet(name init_linguistic_term,
			   float init_lower_support_end,
			   float init_upper_support_end,
			   colourtype init_fs_colour = white,
			   name init_f = "point_list" ) :
	FuzzySetBase(init_linguistic_term, init_lower_support_end,
		     init_upper_support_end, init_fs_colour) {
	function.set_name(init_f);
	match_value = -1.0;
	next=NULL; };

PlimFuzzySet::PlimFuzzySet() :
		FuzzySetBase() {
	function.set_name("point_list");
	match_value = -1.0;
	next=NULL; };

// * * complex methods:

void PlimFuzzySet::show(void) {
	show_FSB();
	cout << "function: "; function.show(); cout << endl;
	cout << "match_value: " << match_value << endl;
	cout << "membership_function: "; msf.show(); cout << endl;
	cout << "instanciated_membership_function: ";
	imsf.show(); cout << endl; };

int PlimFuzzySet::write(ofstream *output) {
	write_FSB(output);
	function.write(output);
	*output << match_value << endl;
	msf.write(output);  *output << endl;
	imsf.write(output); *output << endl;
	return 1; };


int PlimFuzzySet::read(ifstream *input) {
	read_FSB(input);
	function.read(input);
	*input >> match_value;
	msf.read(input);
	imsf.read(input);
	return 1; };

void PlimFuzzySet::calculate_match_value(float crisp_input) {
	float ascent;
	boolean matched = FALSE;

	if (this != NULL) {
	match_value = 0.0;
	if (lower_support_end < crisp_input &&
	    crisp_input < upper_support_end) {
	    // crisp_input lies in support
	    // match_value seting:
	    for (int i=0; i<msf.get_number(); i++)
		if ( msf[i]->get_x() > crisp_input && matched == FALSE ) {
		    // crisp_input lies between msf[i-1] and msf[i]
		    // build line between both points
		    ascent = ( msf[i]->get_y() - msf[i-1]->get_y() ) /
			     ( msf[i]->get_x() - msf[i-1]->get_x() );
		    match_value = msf[i-1]->get_y() + ascent * ( crisp_input
				  - msf[i-1]->get_x() );
		    matched = TRUE; }; }; }; };

void PlimFuzzySet::build_imsf(void) {
	float ascent,new_x;
	int i;

	if (this != NULL) {
	imsf.remove_all();
	if (match_value == 0.0) {
	    imsf.add(lower_support_end,0.0);
	    imsf.add(upper_support_end,0.0); }
	else
	  for (i=0; i<msf.get_number()-1; i++)
	    if (msf[i]->get_y() <= match_value
		&& msf[i+1]->get_y() <= match_value) {
		// both Points below match_value
		if (i<1) imsf.insert(msf[i]->get_x(),msf[i]->get_y());
		imsf.insert(msf[i+1]->get_x(),msf[i+1]->get_y()); }
	    else
	      if (msf[i]->get_y() <= match_value
		  && msf[i+1]->get_y() > match_value) {
		  // ascendent trough match_value
		  if (i<1) imsf.insert(msf[i]->get_x(),msf[i]->get_y());
		  ascent = ( msf[i+1]->get_x() - msf[i]->get_x() ) /
			   ( msf[i+1]->get_y() - msf[i]->get_y() );
		  new_x = msf[i]->get_x() + ascent * ( match_value -
			  msf[i]->get_y() );
		  imsf.insert(new_x,match_value); }
	      else
		if (msf[i]->get_y() > match_value &&
		    msf[i+1]->get_y() <= match_value) {
		    // descendent trough match_value
		    imsf.insert(msf[i+1]->get_x(),msf[i+1]->get_y());
		    ascent = ( msf[i+1]->get_x() - msf[i]->get_x() ) /
			     ( msf[i+1]->get_y() - msf[i]->get_y() );
		    new_x = msf[i]->get_x() + ascent * ( match_value -
			    msf[i]->get_y() );
		    imsf.insert(new_x,match_value); }; }; };


void PlimFuzzySet::add_msf_point(float new_x, float new_y) {
	// check that new_x within support
	if (new_x >= lower_support_end  && new_x <= upper_support_end)
		// check that  0 <= new_y <= 1
		if (new_y >= 0 && new_y <= 1) {
			msf.insert(new_x,new_y);
			msf.clear_lines(); }; };


int PlimFuzzySet::delete_msf_point(int index)  { 
	int   number = msf.get_number();
	if (index == 0 || index >= (number-1)) {
		cerr << "Unable to delete end points!" << endl;
		return -1; }
	msf.remove(index);
	msf.clear_lines(); 
	return 1; };


void PlimFuzzySet::copy(PlimFuzzySet *pfs) {
	copy_FSB        ( pfs->get_FSB());
	match_value     = pfs->match_value;
	msf .copy_PList ( &(pfs->msf));
	imsf.copy_PList ( &(pfs->imsf)); };

FuzzySetBase *PlimFuzzySet::get_FSB(void) {
	FuzzySetBase *fsb = new FuzzySetBase;
	fsb->set_linguistic_term(linguistic_term);
	fsb->set_lower_support_end(lower_support_end);
	fsb->set_upper_support_end(upper_support_end);
	fsb->set_fs_colour(fs_colour);
	return fsb; };

void PlimFuzzySet::set_msf_point_x(int index, float new_x) {
	// check that new_x within support
	// local check that x value is between its neighbours
	if (new_x > upper_support_end) new_x=upper_support_end;
	if (new_x < lower_support_end) new_x=lower_support_end;

	int  number = msf.get_number();
	if (index > 0 && index < number-1) {
		float prev_x = msf[index-1]->get_x();
		float next_x = msf[index+1]->get_x();
		float low_x  = min(prev_x,next_x);
		float high_x = max(prev_x,next_x);
		if (new_x < low_x) new_x = low_x;
		if (new_x > high_x) new_x = high_x; };
	msf[index]->set_x(new_x); 
	msf.clear_horizontal_line();
	msf.clear_vertical_line(); };

void PlimFuzzySet::set_msf_point_y(int index, float new_y) {
	// check that  0 <= new_y <= 1
	// check that no intermediate point is set to 0
	// 	local check that y value is not below its neighbours
	// 	local check that y value is not above its neighbours
	//		except if neighbours hight is max
	// check that first and last are 0
	int   number = msf.get_number();
	float max_y  = msf.get_max_y();

	if (new_y > 1) new_y=1;
	if (new_y < 0) new_y=0;
	if (index > 0 && index < number-1) {
	    if (new_y > 0) {
		float prev_y = msf[index-1]->get_y();
		float next_y = msf[index+1]->get_y();
		float low_y = min(next_y,prev_y);
		if (new_y < low_y) new_y = low_y;
		float high_y = max(next_y,prev_y);
		if (high_y < max_y && new_y > high_y) new_y = high_y;
		msf[index]->set_y(new_y); }; }
	else msf[index]->set_y(0); 
	msf.clear_horizontal_line();
	msf.clear_vertical_line(); };


//
//
//	* * * * * * * * * *  P l i m L i s t  * * *
//
//

// * * destructor:

PlimList::~PlimList(void) {
	while (first != NULL) {
		PlimFuzzySet *e=first;
		first = e->get_next();
		delete e; }; };

// * * complex methods:

void PlimList::show(void) {
	PlimFuzzySet *element=first;
	while (element != NULL) {
		element->show();
		element=element->next; }; };

int PlimList::write(ofstream *output) {
	PlimFuzzySet *element=first;
	*output << "/* #: " << number << " */ " << endl;
	while (element != NULL) {
		element->write(output);
		element=element->next; };
	return 1; };

int PlimList::read(ifstream *input) {
	char instr[50];
	int n;
	*input >> instr; //  /*
	*input >> instr; //  #:
	*input >> n;
	*input >> instr; //  */
	for (int i=0; i<n; i++) {
		add();
		last->read(input); };
	return 1; };

void PlimList::add(void) {
	PlimFuzzySet *element=new PlimFuzzySet;
	if (first == NULL ) { first=element;      last=element; }
	else 		    { last->next=element; last=element; };
	number++; };

void PlimList::add(name init_name,
		 float lower_support_end, float upper_support_end,
		 float lower_dilatation_length, float upper_dilatation_length,
		 colourtype init_colour) {
	PlimFuzzySet *element=new PlimFuzzySet(init_name, lower_support_end,
			      upper_support_end, init_colour);
    //  create msf:
	float x2 = lower_support_end+lower_dilatation_length;
	float x3 = upper_support_end-upper_dilatation_length;
	element->msf.add_PList(lower_support_end, 0.0);
	element->msf.add_PList(x2,1);
	if (x2 != x3) element->msf.add_PList(x3,1);
	element->msf.add_PList(upper_support_end, 0.0);
	if (first == NULL) { first=element;      last=element; }
	else 		   { last->next=element; last=element; };
	number++; };

PlimFuzzySet *PlimList::find(int index) {
	PlimFuzzySet *element=first;
	if (index < number) {
		for (int i=0; i<index; i++) element=element->next;
		return element; }
	else
		return NULL; };
	// for (int j=number; j<=index; j++) add();
	// return last; };

int PlimList::get_fs_index(char *linguistic_term) {
	PlimFuzzySet *element=first;
	int index=0;
	while ( strcmp( element->linguistic_term,linguistic_term)!=0
		&& index < number ) {
		element=element->next;
		index++; };
	if (index<number) return index;
	else return -1; };

float PlimList::get_membership(float x_value, int fs_index) {
	int lower_index,upper_index = 0;
	float ascent;

	if (find(fs_index)->get_lower_support_end() >= x_value) return 0.0;
	if (find(fs_index)->get_upper_support_end() <= x_value) return 0.0;
	for (int i=0; i<number; i++)
		if (find(fs_index)->get_msf_point_x(i) > x_value)
			if ( upper_index == 0 ) {
				upper_index = i;
				lower_index = i-1; };
	// get ascent of line between this two points
	ascent=(find(fs_index)->get_msf_point_y(upper_index)
	      - find(fs_index)->get_msf_point_y(lower_index))
	     / (find(fs_index)->get_msf_point_x(upper_index)
	      - find(fs_index)->get_msf_point_x(lower_index));
	return ( ascent * ( x_value -
	       find(fs_index)->get_msf_point_x(lower_index))
	     + find(fs_index)->get_msf_point_y(lower_index)); };

NList *PlimList::get_names_of_vars(void) {
	NList *n_list = new NList;
	PlimFuzzySet *element=first;
	name n;
	for (int i=0; i<number; i++) {
		strcpy(n,element->get_linguistic_term());
		n_list->add(n);
		element=element->next; };
	return n_list; };

int PlimList::check_match(void) {
	// finds out if cv is already instanciated
	// is TRUE when match value is no longer -1.0
	if (first->get_match_value() >= 0)
		return 0;
	return -1; };

void PlimList::set_all_mvs(float mv) {
	for (int i=0; i<number; i++) {
		find(i)->set_match_value(mv);
		find(i)->build_imsf(); }; };

int PlimList::delete_msf_point(int fs_index, int point_index) {
	return find(fs_index)->delete_msf_point(point_index); };

void PlimList::copy(PlimList *pl) {
	remove_all();
	PlimFuzzySet *element=pl->get_first();
	while (element!=NULL) {
		add();
		last->copy_PFS(element);		
		element=element->get_next(); }; };

int PlimList::put_imsf_into_cover(PList *cover, int imsf_index) {
	PList *union_list = new PList;
	PList *cross_list = new PList;
	PList *imsf       = new PList;
	imsf->copy(find(imsf_index)->get_imsf());
	int cn=cover->get_number();
	int fn=imsf->get_number();

	cover->clear_horizontal_line();
	imsf->clear_horizontal_line();

   // cout << "cover: ";      cover->show(); cout << endl;
   // cout << "imsf: ";       imsf->show(); cout << endl;

	// if any list is empty finish:
	if (cover==NULL || cn <= 2)  {
		if (imsf==NULL || fn <= 2)  {
			cover->remove_all();
	 		delete imsf; delete cross_list; delete union_list;
			return 1; };
		cover->copy(imsf);
	 	delete imsf; delete cross_list; delete union_list;
		return 1; };
	if (imsf==NULL || fn <= 2)  {
		cover->clear_horizontal_line();
	 	delete imsf; delete cross_list; delete union_list;
		return 1; };

   // cout << "cover: ";      cover->show(); cout << endl;
   // cout << "imsf: ";       imsf->show(); cout << endl;

	cn=cover->get_number();
	fn=imsf->get_number();

	// insert all cover-points in union_list:
	Point *element=cover->get_first();
		while (element!=NULL) {
		union_list->insert(element);
		element=element->get_next(); };

	// insert all imsf-points in union_list:
	element=imsf->get_first();
	while (element!=NULL) {
		union_list->insert(element);
		element=element->get_next(); };

	// insert all cross-points into cross_list
	// cout << "cn: " << cn << " fn: " << fn << endl;
	for (int c=1; c<cn; c++) {
	 float cx1=cover->find_x(c-1);
	 float cy1=cover->find_y(c-1);
	 float cx2=cover->find_x(c);
	 float cy2=cover->find_y(c);
	 for (int f=1; f<fn; f++) {
		 float fx1=imsf->find_x(f-1);
		 float fy1=imsf->find_y(f-1);
		 float fx2=imsf->find_x(f);
		 float fy2=imsf->find_y(f);
		 float cx,cy;
		 if (crosspoint(&cx,&cy,cx1,cy1,cx2,cy2,fx1,fy1,fx2,fy2)==1) {
			// cout << "cx: " << cx << " cy: " << cy << endl;
			cross_list->insert(cx,cy); }; }; };
	 // cout << "union_list: "; union_list->show(); cout << endl;

	 // remove all points in union-list that are under the cover-line
	 //						   the imsf-line
	element=cover->get_first();
	while (element!=NULL) {
		if (above(element,imsf)== -1) {
			int pi=union_list->find(element);
			union_list->remove(pi); };
		if (element!=NULL) 
		   element=element->get_next(); };
	element=imsf->get_first();
	while (element!=NULL) {
		if (above(element,cover)== -1) {
			int pi=union_list->find(element);
			union_list->remove(pi); };
		if (element!=NULL) 
		   element=element->get_next(); };

	 // put the union-list with the cross-list together as new cover
	cover->remove_all();
	cover->copy(union_list);
	element=cross_list->get_first();
	while (element!=NULL) {
		cover->insert(element);
		element=element->get_next(); };

	cover->clear_horizontal_line();

	 // cout << "union_list: "; union_list->show(); cout << endl;
	 // cout << "cross_list: "; cross_list->show(); cout << endl;
	 // cout << "cover: ";      cover->show(); cout << endl;
	 delete imsf; delete cross_list; delete union_list;
	 return 1; };

int PlimList::above(float x,float y,float ax,float ay,float bx,float by) {
	// check if the original y-value of the point is
	// higher, lower or equal than the picture of the point

	float pic;
	int   pic_rv=picture(&pic,x,ax,ay,bx,by);
	if (pic_rv<0)  return -3;  // if picture not posible
	if (pic_rv==0) return -2;  // if picture not in x-range of line
	if (y>pic)     return  1;
	if (y==pic)    return  0;
	if (y<pic)     return -1;
		       return -4; };

int PlimList::above(Point *p, PList *l) {
	int ln=l->get_number();
	float x=p->get_x();
	float y=p->get_y();
	for (int i=1; i<ln; i++) {	
		Point *e=l->find(i-1);
		float ax=e->get_x();
		float ay=e->get_y();
		e=l->find(i);
		float bx=e->get_x();
		float by=e->get_y();
		if (x>=ax && x<=bx) {
			int abov_rv=above(x,y,ax,ay,bx,by);
			if (abov_rv > -2) return abov_rv; }; };
	return -2; 	// if no dicision was posible
	};
	
int PlimList::picture(float *pic,float x,float ax,float ay,float bx,float by){
	// get equiation for line ab
	// put x into it

	// line-discussion:
	float dx,dy,asc,y_start;
	boolean vert=FALSE,hor=FALSE;

	dx=bx-ax;	  dy=by-ay;
	if 	(dx==0)   vert=TRUE;
	else if (dy==0) { *pic=ay; hor=TRUE; }
	else            { asc=dy/dx;  y_start=by-asc*bx; };

	if      (vert) 	    return -1;	// no picture posible
	else if (hor) *pic=ay;
	     else     *pic=y_start+asc*x;
	if (x>=ax && x<=bx) return  1;	// if picture in x-range of line
	else		    return  0;  // if picture not in x-range of line
	};


int PlimList::crosspoint(float *cx,float *cy,
			 float ax1,float ay1,float ax2,float ay2,
			 float bx1,float by1,float bx2,float by2) {
	// cross to lines:
	// line discussions for both lines
	// check if there is a crosspoint possible at all!  -> -1
	// calculate crosspoint
	// check if crosspoint is member of both lines      -> 1 v 0

	// line-discussion line a:
	float adx,ady,x_cross,y_cross,aasc,ay_start;
	boolean vert,hor,avert,ahor;
	vert=hor=avert=ahor=FALSE;

	adx=ax2-ax1;	   ady=ay2-ay1;
	if      (adx==0) { x_cross=ax1; avert=TRUE; vert=TRUE; }
	else if (ady==0) { y_cross=ay1; ahor =TRUE; hor =TRUE; }
	else             { aasc=ady/adx; ay_start=ay2-aasc*ax2; };

	// line-discussion line b:
	float bdx,bdy,basc,by_start;
	boolean bvert,bhor;
	bvert=bhor=FALSE;

	bdx=bx2-bx1;	   bdy=by2-by1;
	if      (bdx==0) { x_cross=bx1; bvert=TRUE; vert=TRUE; }
	else if (bdy==0) { y_cross=by1; bhor =TRUE; hor =TRUE; }
	else             { basc=bdy/bdx; by_start=by2-basc*bx2; };

	*cx=0; *cy=-1;
	if ( ahor  && bhor  ) return -1;
	if ( avert && bvert ) return -1;

	if (!(hor && vert)) {
		if (ahor)  x_cross=(y_cross-by_start)/basc;
		if (bhor)  x_cross=(y_cross-ay_start)/aasc;
		if (avert) y_cross=basc*x_cross+by_start;
		if (bvert) y_cross=aasc*x_cross+ay_start;

		if (!hor && !vert) {
			float dy_start,dasc;
			dy_start=by_start-ay_start;
			dasc=basc-aasc;
			if (dasc==0) return -1;
			x_cross=-dy_start/dasc;
			y_cross=aasc*x_cross+ay_start; }; };

	*cx=x_cross;	*cy=y_cross;
	if (((ax1 <= *cx && *cx <= ax2) || (ax2 <= *cx && *cx <= ax1))
	 && ((bx1 <= *cx && *cx <= bx2) || (bx2 <= *cx && *cx <= bx1))
	 && ((ay1 <= *cy && *cy <= ay2) || (ay2 <= *cy && *cy <= ay1))
	 && ((by1 <= *cy && *cy <= by2) || (by2 <= *cy && *cy <= by1)))
		return 1;
	else    return 0; };


void PlimList::set_fs_lower_support_end(int fs_index,float new_value) {
	// new_value has to be less than next point's x value
	float next_x = find(fs_index)->get_msf_point_x(1);
	if (new_value > next_x) new_value = next_x; 
	find(fs_index)->set_msf_point_x(0,new_value);
	find(fs_index)->set_lower_support_end(new_value); };

void PlimList::set_fs_upper_support_end(int fs_index, float new_value) {
	// new_value has to be greater than previous point's x value
	int point_index = find(fs_index)->get_msf_point_number()-1;
	float prev_x = find(fs_index)->get_msf_point_x(point_index-1);
	if (new_value < prev_x) new_value = prev_x; 
	find(fs_index)->set_msf_point_x(point_index,new_value);
	find(fs_index)->set_upper_support_end(new_value); };

void PlimList::remove_all(void) {
	while (first != NULL) {
		PlimFuzzySet *element=first;
		first = element->get_next();
		delete element; };
	first=NULL;
	last=NULL;
	number=0; };

//
//
//	* * * * * * * * * *   C o m b i n e d V a r i a b l e   * * *
//
//

// * * constructors:

CombinedVariable::CombinedVariable(void) : ParameterSetBase(), PlimList() {
	strcpy(cv_name,"empty_variable");
	crisp_value = -1.0;
	next=NULL; };

CombinedVariable::CombinedVariable(ParameterSet *ps, name ncv_name) :
	ParameterSetBase(ps->get_range_lower_end(), ps->get_range_upper_end(),
			 ps->get_median(), ps->get_distribution(),
			 ps->get_dilatation(), ps->get_importance()),
	PlimList() {

     // variable declaration
	int number=ps->get_number_NList();
	float range_length, inner_range_length, distribution_border;
	float lower_border, mean_distance, dilatation_length;
	float lower_support_end[50], upper_support_end[50];
	float lower_dilatation_length, upper_dilatation_length;

     // write what is known allready
	strcpy(cv_name,ncv_name);
	crisp_value = -1.0;
	next=NULL;

     // dilatation range test
	if (dilatation < 0) dilatation=0;
	if (dilatation > 1) dilatation=1;

     // Calculation of the single position parameters 
	range_length = min(range_upper_end - median,
		 median - range_lower_end) * 2.0;
	inner_range_length = range_length * distribution;
	 distribution_border = (range_length - inner_range_length) / 2.0;
	lower_border = median - range_lower_end - range_length / 2.0
		+ distribution_border;
	// upper_border = range_upper_end - median - range_length / 2.0
	//	+ distribution_border;
	mean_distance = inner_range_length / ( number - 1 );
	dilatation_length = mean_distance * dilatation;

     // single supports
	lower_support_end[0]=range_lower_end;
	upper_support_end[0]=range_lower_end + lower_border +
		mean_distance / 2.0 + dilatation_length / 2.0;
	 for ( int i=1; i<number-1; i++ ) {
		lower_support_end[i]=upper_support_end[i-1] -
			dilatation_length;
		upper_support_end[i]=lower_support_end[i] +
			mean_distance + dilatation_length; };
	lower_support_end[number-1]=upper_support_end[number-2] -
			mean_distance / 2.0 - dilatation_length / 2.0;
	upper_support_end[number-1]=range_upper_end;

     // add plims
	CList cl;
	cl.copy_CList(ps->get_CList());
	NList nl;
	nl.copy_NList(ps->get_NList());
	for (int i=0; i<number; i++) {
		lower_dilatation_length=dilatation_length;
		upper_dilatation_length=dilatation_length;
		if (i==0) 	 lower_dilatation_length=0;
		if (i==number-1) upper_dilatation_length=0;
		colourtype ct;
		ct = *((cl.find(i))->get_colour());
		name n;
		strcpy(n,*(nl.find_n(i)));
		add_PFS(n, lower_support_end[i],   upper_support_end[i],
			lower_dilatation_length, upper_dilatation_length, ct);
		}; };

// * * complex methods:

void CombinedVariable::show() {
	cout << "  cv_name: " << cv_name << endl
	     << "crisp_value: " << crisp_value << tab;
	show_PSB();
	show_PL();  cout << endl; };

int CombinedVariable::write(ofstream *output) {
	*output << cv_name << space << crisp_value << space;
	write_PSB(output);  *output << endl;
	write_PL(output);   *output << endl;
	return 1; };

int CombinedVariable::read(ifstream *input) {
	*input >> cv_name >> crisp_value;
	read_PSB(input);
	read_PL(input);
	return 1; };

int CombinedVariable::set_value(float crisp_input) {
	if (crisp_input < range_lower_end ||
	    crisp_input > range_upper_end) {
		cerr << "Crisp input out of range!" << endl;
		return -1; };
	crisp_value = crisp_input;
	int n=number;
	for (int i=0; i<n; i++) {
		PlimFuzzySet *pfs=find(i);
		pfs->calculate_match_value(crisp_input);
		pfs->build_imsf(); 
		pfs=NULL; }; 
	return 1; };

void CombinedVariable::copy(CombinedVariable *cv) {
	copy_PSB      ( cv->get_PSB());
	copy_PlimList ( cv );
	strcpy(cv_name, cv->get_cv_name());
	crisp_value   = cv->get_crisp_value(); };

ParameterSetBase *CombinedVariable::get_PSB(void) {
	ParameterSetBase *psb = new ParameterSetBase;
	psb->set_range_lower_end(range_lower_end);
	psb->set_range_upper_end(range_upper_end);
	psb->set_median         (median);
	psb->set_distribution   (distribution);
	psb->set_dilatation     (dilatation);
	psb->set_importance     (importance);
	return psb; };

void CombinedVariable::copy_PlimList(CombinedVariable *cv) {
	remove_all();
	PlimFuzzySet *element=cv->get_first();
	while (element!=NULL) {
		add_PFS();
		last->copy_PFS(element);		
		element=element->get_next(); }; };


//
//
//	* * * * * * * * * *  C V L i s t  * * *
//
//

// * * destructor:

 CVList::~CVList(void) {
	while (first != NULL) {
		CombinedVariable *cv=first;
		first = cv->get_next();
		delete cv; }; };

// * * complex methods:

void CVList::show(void) {
	CombinedVariable *element=first;
	while (element != NULL) {
		element->show();
		element=element->next; };
	cout << endl; };

int CVList::write(ofstream *output) {
	CombinedVariable *element=first;
	*output << "/* #: " << number << " */" << endl << endl;
	while (element != NULL) {
		element->write(output);
		element=element->next; };
	*output << endl;
	return 1; };

int CVList::read(ifstream *input) {
	char instr[10];
	int n;
	*input >> instr; //  /*
	*input >> instr; //  #:
	*input >> n;
	*input >> instr; //  */
	for (int i=0; i<n; i++) {
		add();
		last->read(input); };
	return 1; };

void CVList::add(void) {
	CombinedVariable *element=new CombinedVariable;
	if (first == NULL ) { first=element;      last=element; }
	else 		    { last->next=element; last=element; };
	number++; };

void CVList::add(CombinedVariable *cv) {
	CombinedVariable *element=new CombinedVariable;
	element->copy(cv);
	if (first == NULL ) { first=element; last=element; number=1; }
	else 		    { last->next=element; last=element; number++; }; };

void CVList::add(CVList *cvl) {
	// adds just if cv   n o t   in CVList!
	// key is the name of the cv and not the address!
	CombinedVariable *element=cvl->get_first();
	while (element != NULL) {
		int index=find_cv_by_name(element);
		if (index < 0) 	add(element);
		element=element->next; }; };

CombinedVariable *CVList::find(int index) {
	CombinedVariable *element=first;
	if (index < number) {
		for (int i=0; i<index; i++)
			element=element->next;
		return element; }
	else 	return NULL; };

CombinedVariable *CVList::find_cv(name cv_name) {
	CombinedVariable *cv=first;
	while (cv!=NULL) {
		if (strcmp(cv->get_cv_name(),cv_name)==0)
			return cv;
		cv=cv->get_next(); };
	return NULL; };

int CVList::find_cv_by_pointer(CombinedVariable *cv) {
	//compare pointers
	//dangerouse because of copy in any add function
	int i=0,n=number;
	if (n<=0) return -1;
	while (i<n && find(i)!=cv) i++;
	if (i<n) return i;
	return -1; };

int CVList::find_cv_by_name(CombinedVariable *cv) {
	// compare names of cvs
	int n=number,index=-1;
	name n1,n2;
	strcpy(n1,cv->get_cv_name());
	for (int i=0; i<n; i++) {
		strcpy(n2,find(i)->get_cv_name());	
		if (strcmp(n1,n2)==0) { index=i; break; }; };
	return index; };

void CVList::remove(int index) {
	CombinedVariable *element;
	CombinedVariable *to_delete;
	if (number==0) { first=NULL; last=NULL; } 
	else { if (index==0) { 
			to_delete=first;
			first=first->next; }
		else { 	element=find(index-1);
	       		to_delete=element->next;
	       		element->next=element->next->next; };
		number--;
		delete to_delete; }; };

void CVList::remove_all(void) {
	while (first != NULL) {
		CombinedVariable *cv=first;
		first = cv->get_next();
		delete cv; }; 
	number=0;
	first=NULL;
	last=NULL; };


//
//
//	* * * * * *   o t h e r e   g l o b a l  f u n t i o n s   * * *
//
//

float min(float a, float b) {
	if ( a < b ) return a;
	return b; }

float max(float a, float b) {
	if ( a > b ) return a;
	return b; }

void wait_for_key_press(void) {
	cout << "-> press return to continue! ";
	getchar();
	cout << endl; };
