/************************************************************************
 * This library is free software; you can redistribute it and/or	*
 * modify it under the terms of the GNU Library General Public		*
 * License as published by the Free Software Foundation; either		*
 * version 2 of the License, or (at your option) any later version.	*
 * 									*
 * This library is distributed in the hope that it will be useful,	*
 * but WITHOUT ANY WARRANTY; without even the implied warranty of	*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU	*
 * Library General Public License for more details.			*
 * 									*
 * You should have received a copy of the GNU Library General Public	*
 * License along with this library; if not, write to the		*
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,		*
 * Boston, MA  02111-1307, USA.						*
 * 									*
 * The GNU Library General Public License file is included in		*
 * Online-Documentation							*
 *									*
 * GNU's Homepage - http://www.gnu.ai.mit.edu/home.html			*
 *									*
 * Questions & comments referring to *FLIP++: please contact us:	*
 *									*
 * Snail-Mail:								*
 *	Dr. Wolfgang SLANY						*
 *	Database and Expert System Group (E184-2)     			*
 *	Institute of Information Systems               			*
 *	Vienna University of Technology					*
 *	Paniglgasse 16, A-1040 Vienna, AUSTRIA				*
 * Telephone:								*
 *	+43-1-58801-6123     (area-code 1 maybe changed to 201 in 1998) *
 * FAX:									*
 *	+43-1-5055304	     (area-code 1 maybe changed to 201 in 1998) *
 * eMail:								*
 *	sflip@dbai.tuwien.ac.at						*
 * WWW:									*
 *	http://www.dbai.tuwien.ac.at/proj/StarFLIP/			*
 *                                                                      *
 * $Id: flipdemo.cpp,v 1.4 1997/06/24 17:30:02 reichel Exp $ *
 ***********************************************************************/


//
//     ------------------------------------
//	F L I P D E M O . C C  ( . C P P )
//     ------------------------------------
//

#include "flip.h"

main() {
cout.precision(3);
cout << endl << endl << endl;
cout << "                -----------------------------        "<< endl;
cout << "	       ---------------------------------      "<< endl;
cout << "	      |                                 |     "<< endl;
cout << "	    | |       F L I P  -  D E M O       | |   "<< endl;
cout << "	      |                                 |     "<< endl;
cout << "	       ---------------------------------      "<< endl;
cout << "	         -----------------------------        "<< endl;

// *** declare and allocate io-streams

ifstream *input = new ifstream;
ofstream *output = new ofstream;

// *** declare errorstring

static char error[100]=" ";

cout << endl << endl
     << tab << tab << "*************************************" 
     << endl << endl
     << tab << tab << "c r e a t e   P a r a m e t e r S e t" 
     << endl << endl
     << tab << tab << "*************************************" 
     << endl;

cout << endl << "// * * create CList:" << endl;
	CList *cl_outdoor  = new CList(4);
	CList *cl_indoor   = new CList(3);
	CList *cl_heating  = new CList(3);

cout << endl << "// * * create NList:" << endl;
// NList needet as paramter to create a ParameterSet
// !! the names are  n o t  allowed to include spaces !!
	NList *nl_outdoor = new NList;
	NList *nl_indoor  = new NList;
	NList *nl_heating = new NList;

cout << endl << "// * * adapt NList:" << endl;
	nl_outdoor->add("icy");
	nl_outdoor->add("cool");
	nl_outdoor->add("mild");
	nl_outdoor->add("hot");
	nl_indoor->add("cold");
	nl_indoor->add("medium");
	nl_indoor->add("warm");
	nl_heating->add("no");
	nl_heating->add("half");
	nl_heating->add("full");

cout << endl << "// * * create ParameterSet:" << endl;
	ParameterSet *ps_outdoor,*ps_indoor,*ps_heating;
	ps_outdoor = new ParameterSet(
		"ps_outdoor", -50, 40, 0, 0.9, 0.7, 1, cl_outdoor, nl_outdoor);
	ps_indoor  = new ParameterSet(
		"ps_indoor", 0, 40, 15, 0.9, 0.7, 0.9, cl_indoor, nl_indoor);
	ps_heating = new ParameterSet(
		"ps_heating", 0, 100, 50, 0.9, 0.8, 1, cl_heating, nl_heating);

if (strcmp(error," ")!=0) cout << endl << " ### error: " << error << endl;

cout << endl << endl
     << tab << tab << "*********************************************"
     << endl << endl
     << tab << tab << "c r e a t e   C o m b i n e d V a r i a b l e"
     << endl << endl
     << tab << tab << "*********************************************"
     << endl;

cout << endl << "// * * construct CombinedVariable:" << endl;
	CombinedVariable *cv_outdoor = 
		new CombinedVariable(ps_outdoor,"outdoor");
	CombinedVariable *cv_indoor = 
		new CombinedVariable(ps_indoor,"indoor");
	CombinedVariable *cv_heating = 
		new CombinedVariable(ps_heating,"heating");

cout << endl << "// * * set_value" << endl;
	cv_outdoor->set_value(7);
	cv_indoor->set_value(17);

if (strcmp(error," ")!=0) cout << endl << " ### error: " << error << endl;

cout << endl << endl
     << tab << tab << "*************************" 
     << endl << endl
     << tab << tab << "R u l e S e t  -  t e s t" 
     << endl << endl
     << tab << tab << "*************************"
     << endl;

cout << endl << "// * * create RuleSet:" << endl;
	RuleSet *rule_set = new RuleSet();

cout << endl << "// * * test OperatorSet - set:" << endl;
	rule_set->get_operator_set()->set_inference_operator(maximum);

cout << endl << "// * *  test PSList add" << endl;
	rule_set->get_ps_list()->add(ps_outdoor);
	rule_set->get_ps_list()->add(ps_indoor);
	rule_set->get_ps_list()->add(ps_heating);

cout << endl << "// * *  test RuleSet add" << endl;
	rule_set->add_ComplexRule(
		cv_outdoor, "icy",  cv_indoor, "cold",
		cv_heating, "full", AND, 1);
	rule_set->add_ComplexRule(
		cv_outdoor, "icy",  cv_indoor, "medium",
		cv_heating, "full", AND, 0.9);
	rule_set->add_ComplexRule(
		cv_outdoor, "icy",  cv_indoor, "warm",
		cv_heating, "half", AND, 1);
	rule_set->add_ComplexRule(
		cv_outdoor, "cool",  cv_indoor, "cold",
		cv_heating, "full", AND, 0.8);
	rule_set->add_ComplexRule(
		cv_outdoor, "cool",  cv_indoor, "medium",
		cv_heating, "half", AND, 1);
	rule_set->add_ComplexRule(
		cv_outdoor, "cool",  cv_indoor, "warm",
		cv_heating, "half", AND, 0.7);
	rule_set->add_ComplexRule(
		cv_outdoor, "mild",  cv_indoor, "cold",
		cv_heating, "half", AND, 1);
	rule_set->add_ComplexRule(
		cv_outdoor, "mild",  cv_indoor, "medium",
		cv_heating, "half", AND, 0.8);

cout << endl << "// * * show RuleSet:" << endl;
	rule_set->show();

cout << endl << "// * *  show CVList" << endl;
	rule_set->get_cv_list()->show();


if (strcmp(error," ")!=0) cout << endl << " ### error: " << error << endl;

cout << endl << endl
     << tab << tab << "***************************" 
     << endl << endl
     << tab << tab << "e v a l u a t e  -  t e s t" 
     << endl << endl
     << tab << tab << "***************************" 
     << endl;

cout << endl << "// * * evaluate cv_heating:" << endl;
	rule_set->evaluate(cv_heating);

if (strcmp(error," ")!=0) cout << endl << " ### error: " << error << endl;

cout << endl << endl
     << tab << tab << "***************************" 
     << endl << endl
     << tab << tab << "H A N D L I N G  -  T E S T" 
     << endl << endl
     << tab << tab << "***************************" 
     << endl;

cout << endl << "// * * write RuleSet:" << endl;
	output->open("fliprule.dat");
	rule_set->write(output);
	output->close();

cout << endl << "// * * create 2nd RuleSet:" << endl;
	RuleSet *rule_set2 = new RuleSet();

cout << endl << "// * * construct new CombinedVariable:" << endl;
	CList *cl_fuel_flow  = new CList(3);
	NList *nl_fuel_flow = new NList;
	nl_fuel_flow->add("zero");
	nl_fuel_flow->add("low");
	nl_fuel_flow->add("high");
	ParameterSet *ps_fuel_flow = new ParameterSet("ps_fuel_flow", 
		0, 50, 20, 0.9, 0.9, 1, cl_fuel_flow, nl_fuel_flow);
	rule_set2->get_ps_list()->add(ps_fuel_flow);
	CombinedVariable *cv_fuel_flow = 
		new CombinedVariable(ps_fuel_flow,"fuel_flow");

cout << endl << "// * *  RuleSet add new Rules" << endl;
	// cv_heating is here still known from above
	rule_set2->add_BasicRule( cv_heating, "no",   cv_fuel_flow, "zero", 1);
	rule_set2->add_BasicRule( cv_heating, "half", cv_fuel_flow, "low",  1);
	rule_set2->add_BasicRule( cv_heating, "full", cv_fuel_flow, "high", 1);
	
cout << endl << "// * * add old RuleSet:" << endl;
	delete rule_set;
	RuleSet *rule_set_temp = new RuleSet();
	input->open("fliprule.dat");
	rule_set_temp->read(input);
	input->close();
	rule_set2->get_ps_list()->add(rule_set_temp->get_ps_list());
	rule_set2->get_cv_list()->add(rule_set_temp->get_cv_list());
	rule_set2->add_all_RuleBlocks(rule_set_temp);
	delete rule_set_temp;

cout << endl << "// * * evaluate cv_fuel_flow:" << endl;
	rule_set2->evaluate(cv_fuel_flow);

cout << endl << "// * * show combined RuleSet:" << endl;
	rule_set2->show();

cout << endl << "// * * show combined OperatorSet:" << endl;
	rule_set2->get_operator_set()->show();

cout << endl << "// * *  show combined CVList:" << endl;
	rule_set2->get_cv_list()->show();

cout << endl << "// * * show combined PSList:" << endl;
	rule_set2->get_ps_list()->show();

cout << endl << "// * * remove_cv_and_its_approaches(indoor):" << endl;
	rule_set2->remove_cv_and_its_approaches(cv_indoor);

cout << endl << "// * * show combined RuleSet after removing indoor:" << endl;
	rule_set2->show();

cout << endl << "// * *  show combined CVList after removing indoor:" << endl;
	rule_set2->get_cv_list()->show();

if (strcmp(error," ")!=0) cout << endl << " ### error: " << error << endl;

	}; //   end of flipdemo.cpp
