/************************************************************************
 * This library is free software; you can redistribute it and/or	*
 * modify it under the terms of the GNU Library General Public		*
 * License as published by the Free Software Foundation; either		*
 * version 2 of the License, or (at your option) any later version.	*
 * 									*
 * This library is distributed in the hope that it will be useful,	*
 * but WITHOUT ANY WARRANTY; without even the implied warranty of	*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU	*
 * Library General Public License for more details.			*
 * 									*
 * You should have received a copy of the GNU Library General Public	*
 * License along with this library; if not, write to the		*
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,		*
 * Boston, MA  02111-1307, USA.						*
 * 									*
 * The GNU Library General Public License file is included in		*
 * Online-Documentation							*
 *									*
 * GNU's Homepage - http://www.gnu.ai.mit.edu/home.html			*
 *									*
 * Questions & comments referring to *FLIP++: please contact us:	*
 *									*
 * Snail-Mail:								*
 *	Dr. Wolfgang SLANY						*
 *	Database and Expert System Group (E184-2)     			*
 *	Institute of Information Systems               			*
 *	Vienna University of Technology					*
 *	Paniglgasse 16, A-1040 Vienna, AUSTRIA				*
 * Telephone:								*
 *	+43-1-58801-6123     (area-code 1 maybe changed to 201 in 1998) *
 * FAX:									*
 *	+43-1-5055304	     (area-code 1 maybe changed to 201 in 1998) *
 * eMail:								*
 *	sflip@dbai.tuwien.ac.at						*
 * WWW:									*
 *	http://www.dbai.tuwien.ac.at/proj/StarFLIP/			*
 *                                                                      *
 * $Id: flipoper.cpp,v 1.4 1997/06/24 17:30:03 reichel Exp $ *
 ***********************************************************************/


//
// $Id: flipoper.cpp,v 1.4 1997/06/24 17:30:03 reichel Exp $
// 
//
//     ------------------------------------
//	F L I P O P E R . C C  ( . C P P )
//     ------------------------------------
//

#include "flipoper.h"

//
//
//	* * * * * * * * * *   i n t e r f a c e    * * *
//
//

float join_mv_with_importance(float mv,float imp,OperatorSet *operator_set) {
	float g=operator_set->get_gamma();
	switch (*(operator_set->get_importance_operator())) {
	  case minimum	   : TNorm t;   return t.min(mv,imp);
	  case product	   : 		return t.prod(mv,imp);
	  case maximum	   : TCoNorm s; return s.max(mv,imp);
	  case sum	   : 		return s.addition(mv,imp);
	  case fuzzy_and   : CompOp c;  return c.fuzzy_and(mv,imp,g);
	  case fuzzy_or	   :            return c.fuzzy_or(mv,imp,g);
	  case ass_comp_and:            return c.ass_comp_and(mv,imp);
	  case ass_comp_or :            return c.ass_comp_or(mv,imp);
	  default	   : return mv; }; };

float aggregate(FList *mvs, concat_operator *agg_op,
		OperatorSet *operator_set) {
	fuzzy_operator *op;
	if      (*agg_op==c_and || *agg_op==AND)
		op=operator_set->get_and_operator();
	else if (*agg_op==c_or || *agg_op==OR)
		op=operator_set->get_or_operator();
	else    return -1;
	float g=operator_set->get_gamma();
	switch (*op) {
	  case minimum	   : TNorm t;   return t.min(mvs);
	  case product	   : 		return t.prod(mvs);
	  case maximum	   : TCoNorm s; return s.max(mvs);
	  case sum	   : 		return s.addition(mvs);
	  case fuzzy_and   : CompOp c;  return c.fuzzy_and(mvs,g);
	  case fuzzy_or	   :            return c.fuzzy_or(mvs,g);
	  case ass_comp_and:            return c.ass_comp_and(mvs);
	  case ass_comp_or :            return c.ass_comp_or(mvs);
	  default	   : return -1; }; };

float certify(float oamv, float certainty, OperatorSet *operator_set) {
	float g=operator_set->get_gamma();
	switch (*(operator_set->get_certainty_operator())) {
	  case minimum	   : TNorm t;   return t.min(oamv,certainty);
	  case product	   : 		return t.prod(oamv,certainty);
	  case maximum	   : TCoNorm s; return s.max(oamv,certainty);
	  case sum	   : 		return s.addition(oamv,certainty);
	  case fuzzy_and   : CompOp c;  return c.fuzzy_and(oamv,certainty,g);
	  case fuzzy_or	   :            return c.fuzzy_or(oamv,certainty,g);
	  case ass_comp_and:            return c.ass_comp_and(oamv,certainty);
	  case ass_comp_or :            return c.ass_comp_or(oamv,certainty);
	  default	   : return -1; }; };

void inference(float oacmv, PlimFuzzySet *fs, OperatorSet *operator_set) {
	// inference with old match_value of this fuzzy_set
	float g=operator_set->get_gamma();
	switch (*(operator_set->get_inference_operator())) {
	  case minimum	   : TNorm t;   t.min(oacmv,fs);          break;
	  case product	   : 		t.prod(oacmv,fs);	  break;
	  case maximum	   : TCoNorm s; s.max(oacmv,fs);          break;
	  case sum	   : 		s.addition(oacmv,fs);	  break;
	  case fuzzy_and   : CompOp c;  c.fuzzy_and(oacmv,fs,g);  break;
	  case fuzzy_or	   :            c.fuzzy_or(oacmv,fs,g);   break;
	  case ass_comp_and:            c.ass_comp_and(oacmv,fs); break;
	  case ass_comp_or :            c.ass_comp_or(oacmv,fs);  break; 
	  default:	;}; };

float accumulate(CombinedVariable *cv, OperatorSet *operator_set) {
	float crisp_value = -1;
	switch (*(operator_set->get_defuzzy_operator())) {
		case cog    :	 DefOp d; crisp_value = d.cog(cv); break;
		case fimax  :	 	  crisp_value = d.first_max(cv); break;
		case mom    :	 	  crisp_value = d.mom(cv); break;
		default:	 return -1; };
	cv->set_crisp_value(crisp_value);
	return crisp_value; };

int calculate(CombinedVariable *cvr, 
	      CombinedVariable *cva, CombinedVariable *cvb, 
	      concat_operator *agg_op, OperatorSet *os) {
	// combines 2 CombinedVariables if they have the same number 
	// of terms
	// sets crisp_value to -1 (does no defuzzification)

	if (cvr == NULL) {
		cerr << "Cannot do calclutions for Null pointer!" << endl;
		return -1; };
	if (cva == NULL && cvb == NULL) {
		cerr << "Nothing to calclute!" << endl;
		return -1; };
	if (cva == NULL) {
		cvr->copy_PSB(cvb->get_PSB());
		cvr->copy_PlimList(cvb);
		cvr->set_crisp_value(-1);
		return 1; };
	if (cvb == NULL) {
		cvr->copy_PSB(cva->get_PSB());
		cvr->copy_PlimList(cva);
		cvr->set_crisp_value(-1);
		return 1; };


	int na = cva->get_number();
	int nb = cvb->get_number();
	int nr = cvr->get_number();
	if (nr != na || nr != nb ) {
		cerr << "Variables have different number of terms!" << endl;
		return -1; };

	float impa  = cva->get_importance();
	float impb  = cvb->get_importance();
	PlimFuzzySet *fs;
	float fs_mva,fs_mvb,mva,mvb,oamv;
	FList *mvs = new FList;
	
	for (int i=0; i<nr; i++) {
		mvs->remove_all();
		fs_mva = cva->get_fs_match_value(i);
		fs_mvb = cvb->get_fs_match_value(i);
		mva = join_mv_with_importance(fs_mva,impa,os);
		mvb = join_mv_with_importance(fs_mvb,impb,os);
		mvs->add(mva);
		mvs->add(mvb);
		cvr->find(i)->set_match_value(0);
		cvr->find(i)->build_imsf(); 
		oamv = aggregate(mvs,agg_op,os);
		fs   = cvr->find(i);
		inference(oamv, fs, os); };

	cvr->set_crisp_value(-1);
	delete mvs;
	return 1; };

//
//
//	* * * * * * * * * *   F u z z y O p e r a t o r   * * *
//
//

// * * complex methods:

void FuzzyOperator::show(void) {
	name op_str;
	switch (oper) {
		case minimum     : strcpy(op_str,"minimum");      break;
		case maximum     : strcpy(op_str,"maximum");      break;
		case fuzzy_and   : strcpy(op_str,"fuzzy_and");    break;
		case fuzzy_or    : strcpy(op_str,"fuzzy_or");     break;
		case ass_comp_and: strcpy(op_str,"ass_comp_and"); break;
		case ass_comp_or : strcpy(op_str,"ass_comp_or");  break;
		case product     : strcpy(op_str,"product");	  break;
		case sum	 : strcpy(op_str,"sum");  	  break;
		default          : strcpy(op_str,"no_fop"); };
	cout << op_str << space; };

int FuzzyOperator::write(ofstream *output) {
	name op_str;
	switch (oper) {
		case minimum     : strcpy(op_str,"minimum");      break;
		case maximum     : strcpy(op_str,"maximum");      break;
		case fuzzy_and   : strcpy(op_str,"fuzzy_and");    break;
		case fuzzy_or    : strcpy(op_str,"fuzzy_or");     break;
		case ass_comp_and: strcpy(op_str,"ass_comp_and"); break;
		case ass_comp_or : strcpy(op_str,"ass_comp_or");  break;
		case product	 : strcpy(op_str,"product"); 	  break;
		case sum 	 : strcpy(op_str,"sum");  	  break;
		default          : strcpy(op_str,"no_fop"); };
	*output << op_str << space;
	return 1; };

int FuzzyOperator::read(ifstream *input) {
	name op_str;
	*input >> op_str;
	oper=no_fop;
	if (strcmp(op_str,"minimum")==0)	oper=minimum;
	if (strcmp(op_str,"maximum")==0) 	oper=maximum;
	if (strcmp(op_str,"fuzzy_and")==0) 	oper=fuzzy_and;
	if (strcmp(op_str,"fuzzy_or")==0) 	oper=fuzzy_or;
	if (strcmp(op_str,"ass_comp_and")==0) 	oper=ass_comp_and;
	if (strcmp(op_str,"ass_comp_or")==0) 	oper=ass_comp_or;
	if (strcmp(op_str,"product")==0) 	oper=product;
	if (strcmp(op_str,"sum")==0) 		oper=sum;
	return 1; };

//
//
//	* * * * * * * * * *   D e F u z z y O p e r a t o r   * * *
//
//

// * * complex methods:

void DeFuzzyOperator::show(void) {
	name op_str;
	switch (oper) {
		case cog:	  strcpy(op_str,"cog"); break;
		case fimax:	  strcpy(op_str,"fimax"); break;
		case mom:	  strcpy(op_str,"mom"); break;
		default:          strcpy(op_str,"no_dfop"); };
	cout << op_str << space; };

int DeFuzzyOperator::write(ofstream *output) {
	name op_str;
	switch (oper) {
		case cog:	strcpy(op_str,"cog"); break;
		case fimax:	strcpy(op_str,"fimax"); break;
		case mom:	strcpy(op_str,"mom"); break;
		default:        strcpy(op_str,"no_dfop"); };
	*output << op_str << space;
	return 1; };

int DeFuzzyOperator::read(ifstream *input) {
	name op_str;
	*input >> op_str;
	oper=no_dfop;
	if      (strcmp(op_str,"cog")==0)	oper=cog;
	else if (strcmp(op_str,"fimax")==0)	oper=fimax;
	else if (strcmp(op_str,"mom")==0)	oper=mom;
	return 1; };
//
//
//	* * * * * * * * * *   C o n c a t O p e r a t o r   * * *
//
//

// * * complex methods:

void ConcatOperator::show(void) {
	name op_str;
	switch (oper) {
		case c_and:	  strcpy(op_str,"AND"); break;
		case c_or:	  strcpy(op_str,"OR"); break;
		case AND:	  strcpy(op_str,"AND"); break;
		case OR:	  strcpy(op_str,"OR"); break;
		default:          strcpy(op_str,"no_cop"); };
	cout << op_str << space; };

int ConcatOperator::write(ofstream *output) {
	name op_str;
	switch (oper) {
		case c_and:	  strcpy(op_str,"AND"); break;
		case c_or:	  strcpy(op_str,"OR"); break;
		case AND:	  strcpy(op_str,"AND"); break;
		case OR:	  strcpy(op_str,"OR"); break;
		default:          strcpy(op_str,"no_cop"); };
	*output << op_str << space;
	return 1; };

int ConcatOperator::read(ifstream *input) {
	name op_str;
	*input >> op_str;
	oper=no_cop;
	if (strcmp(op_str,"AND")==0)	oper=AND;
	if (strcmp(op_str,"OR")==0) 	oper=OR;
	if (strcmp(op_str,"c_and")==0)	oper=AND;
	if (strcmp(op_str,"c_or")==0) 	oper=OR;
	return 1; };

int  ConcatOperator::compare(ConcatOperator *c1, ConcatOperator *c2) {
	if (*c1->get_oper() == *c2->get_oper())	return 1;
	return 0; };


//
//
//	* * * * * * * * * *   t - n o r m   * * *
//
//

float TNorm::min(FList *mvs) {
	float min=1.0;
	int   i=0;
	while (mvs->find(i) != NULL) {
		float v=mvs->find_v(i);
		if (v<min) min=v;
		i++; };
	return min; };

float TNorm::min(float oamv,  float cer) {
	 if ( oamv < cer ) return oamv;
	 else return cer; };

void  TNorm::min(float aocmv, PlimFuzzySet *fs) {
	float mv_new,mv_old;
	TCoNorm s;
	mv_old = fs->get_match_value();
	if (mv_old>0 && aocmv>0) mv_new = min(aocmv,mv_old);
	else mv_new = s.max(aocmv,mv_old);
	fs->set_match_value(mv_new);
	fs->build_imsf(); };

float TNorm::prod(FList *mvs) {
	float prod=1.0;
	int   i=0;
	while (mvs->find(i) != NULL) {
		float v=mvs->find_v(i);
		prod *= v;
		i++; };
	return prod; };

float TNorm::prod(float oamv,  float cer) {
	 return cer*oamv; };

void  TNorm::prod(float aocmv, PlimFuzzySet *fs) {
	float mv_new,mv_old;
	TCoNorm s;
	mv_old = fs->get_match_value();
	if (mv_old>0 && aocmv>0) mv_new = prod(aocmv,mv_old);
	else mv_new = s.max(aocmv,mv_old);
	fs->set_match_value(mv_new);
	fs->build_imsf(); };

//
//
//	* * * * * * * * * *   t - c o - n o r m   * * *
//
//

float TCoNorm::max(FList *mvs) {
	float max=0.0;
	int   i=0;
	while (mvs->find(i)!=NULL) {
		float v=mvs->find_v(i);
		if (v>max) max=v;
		i++; };
	return max; };

float TCoNorm::max(float oamv,  float cer) {
	 if ( oamv > cer ) return oamv;
	 else return cer; };

void  TCoNorm::max(float aocmv, PlimFuzzySet *fs) {
	float mv=max(aocmv,fs->get_match_value());
	fs->set_match_value(mv);
	fs->build_imsf(); };

float TCoNorm::addition(FList *mvs) {
	float sum=0.0;
	int   i=0;
	while (mvs->find(i)!=NULL) {
		float v=mvs->find_v(i);
		sum += v;
		i++; };
	if (sum > 1) sum = 1;
	return sum; };

float TCoNorm::addition(float oamv,  float cer) {
	float sum=oamv+cer;
	if (sum > 1) sum = 1;
	return sum; };

void  TCoNorm::addition(float aocmv, PlimFuzzySet *fs) {
	float mv=addition(aocmv,fs->get_match_value());
	fs->set_match_value(mv);
	fs->build_imsf(); };

//
//
//	* * * * *    c o m p a n s a t o r y   -   o p e r a t o r    * * *
//
//

float CompOp::ass_comp_and(FList *mvs) {
	int n=mvs->get_number();
	float ret_val=mvs->find_v(0);
	if (ret_val<0) ret_val=0;
	for (int i=0; i<n; i++) {
		float mv=mvs->find_v(i);
		if (mv<0) mv=0;
		ret_val=ass_comp_and(ret_val,mvs->find_v(i)); };
	return ret_val; };

float CompOp::ass_comp_and(float x, float y) {
	if (x<0 || x>1 || y<0 || y>1) return -1;
	if (x==0 || x==1 || y==0 || y==1) return 0;
  	double fx=tan(pi/2*double(x));
  	double fy=tan(pi/2*double(y));
  	double c=atan(fx*fy);
  	return float(c); };

void  CompOp::ass_comp_and(float aocmv, PlimFuzzySet *fs) {
	float mv=ass_comp_and(aocmv,fs->get_match_value());
	fs->set_match_value(mv);
	fs->build_imsf(); };

float CompOp::ass_comp_or(FList *mvs) {
	int n=mvs->get_number();
	float ret_val=mvs->find_v(0);
	if (ret_val<0) ret_val=0;
	for (int i=0; i<n; i++) {
		float mv=mvs->find_v(i);
		if (mv<0) mv=0;
		ret_val=ass_comp_or(ret_val,mvs->find_v(i)); };
	return ret_val; };

float CompOp::ass_comp_or(float x, float y) {
	if (x<0 || x>1 || y<0 || y>1) return -1;
	if (x==0 || x==1 || y==0 || y==1) return 0;
   	double fx=1/tan(pi/2*double(x));
  	double fy=1/tan(pi/2*double(y));
  	double c=1-atan(fx*fy);
  	return float(c); };

void  CompOp::ass_comp_or(float aocmv, PlimFuzzySet *fs) {
	float mv=ass_comp_or(aocmv,fs->get_match_value());
	fs->set_match_value(mv);
	fs->build_imsf(); };

float CompOp::fuzzy_and(FList *mvs, float gamma) {
	if (gamma<0 && gamma>1) { cerr << "gamma not in [0,1]!" << endl; 
				  return -1; };
	TNorm t;
	float m=t.min(mvs);
	int n=mvs->get_number();
	float sum=0;
	for (int i=0; i<n; i++) sum+=mvs->find_v(i);
	float algebr_mean=sum/n;
	float ret_val=gamma*m+(1-gamma)*algebr_mean;
	return ret_val; };

float CompOp::fuzzy_and(float oamv,  float cer, float gamma) {
	if (gamma<0 && gamma>1) { cerr << "gamma not in [0,1]!" << endl; 
				  return -1; };
	TNorm t;
	float m=t.min(oamv,cer);
	float algebr_mean=(oamv+cer)/2;
	float ret_val=gamma*m+(1-gamma)*algebr_mean;
	return ret_val; };

void  CompOp::fuzzy_and(float aocmv, PlimFuzzySet *fs, float gamma) {
	if (gamma<0 && gamma>1) cerr << "gamma not in [0,1]!" << endl; 
	else {
		float mv=fuzzy_and(aocmv,fs->get_match_value(),gamma);
		fs->set_match_value(mv);
		fs->build_imsf(); }; };

float CompOp::fuzzy_or(FList *mvs, float gamma) {
	if (gamma<0 && gamma>1) {  cerr << "gamma not in [0,1]!" << endl; 
				  return -1; };
	TCoNorm s;
	float m=s.max(mvs);
	int n=mvs->get_number();
	float sum=0;
	for (int i=0; i<n; i++) sum+=mvs->find_v(i);
	float algebr_mean=sum/n;
	float ret_val=gamma*m+(1-gamma)*algebr_mean;
	return ret_val; };

float CompOp::fuzzy_or(float oamv,  float cer, float gamma) {
	if (gamma<0 && gamma>1) { cerr << "gamma not in [0,1]!" << endl; 
				  return -1; };
	TCoNorm s;
	float m=s.max(oamv,cer);
	float algebr_mean=(oamv+cer)/2;
	float ret_val=gamma*m+(1-gamma)*algebr_mean;
	return ret_val; };

void  CompOp::fuzzy_or(float aocmv, PlimFuzzySet *fs, float gamma) {
	if (gamma<0 && gamma>1) cerr << "gamma not in [0,1]!" << endl; 
	else {
		float mv=fuzzy_or(aocmv,fs->get_match_value(),gamma);
		fs->set_match_value(mv);
		fs->build_imsf(); }; };

//
//
//	* * * * * * * * * *   d e f u z z y   * * *
//
//

float DefOp::cog(CombinedVariable *cv) {
	// need to calculate cog from all imsf >0
	PList *cover=new PList;  // list of all relevent Points
	for (int imsf_index=0; imsf_index<cv->get_number(); imsf_index++)
		{ cv->put_imsf_into_cover(cover,imsf_index); };
  	// cout << endl << "cover: "; cover->show(); cout << endl;
  	// cout << "cover_number: " << cover->get_number(); cout << endl;
	float retval = integrate(cover);
	delete cover;
	return retval; };

float DefOp::first_max(CombinedVariable *cv) {
	// need to get max_y from all imsf >0
	PlimFuzzySet *element=cv->get_first();
	PList *pl;
	float y,y_x,max=0;

	while (element!=NULL) {
		pl = element->get_imsf();
		y = pl->get_max_y();
		if (y>max) max=y;
		element=element->get_next(); }; 
	y_x = pl->get_first_x(y);
	return y_x; };


float DefOp::mom(CombinedVariable *cv) {
	// need to get max_y from all imsf >0
	PlimFuzzySet *element=cv->get_first();
	PList *pl;
	FList *fl = new FList;
	float y,mom_x,max;
	
	max = first_max(cv);
	while (element!=NULL) {
		pl = element->get_imsf();
		y = pl->get_max_y();
		if (y >= max) pl->get_x(fl,y);
		element=element->get_next(); }; 
	mom_x = fl->mean();
	return mom_x; };

float DefOp::integrate(PList *cpl) {
	// formula:
	// cog(cpl)=Moment(cpl)/Area(cpl)
	if (cpl==NULL || cpl->get_number()<2) return -1;
	float M=0,A=0,Mi,Ai;
	int n=cpl->get_number();
	for (int i=1; i<n; i++) {
		float x1=cpl->find(i-1)->get_x();
		float x2=cpl->find(i)->get_x();
		float y1=cpl->find(i-1)->get_y();
		float y2=cpl->find(i)->get_y();
		float dx=x2-x1;
		if ((dx>0) && (y1>0 || y2>0)) {
			Ai=(y1+y2)*dx;
			Mi=(y1*x1+y2*x2)*dx;
			A+=Ai;
			M+=Mi; }; };
	if (A<=0) return -1;
	float cog=M/A;
	return cog; };

//
//
//	* * * * * * * * * *   O p e r a t o r S e t  * * *
//
//

// * * complex methods:

void OperatorSet::show(void) {
	cout << "importance_operator: "; importance_operator.show();
	cout << endl;
	cout << "and_operator: "; and_operator.show(); cout << endl;
	cout << "or_operator: "; or_operator.show(); cout << endl;
	cout << "certainty_operator: "; certainty_operator.show();
	cout << endl;
	cout << "inference_operator: "; inference_operator.show();
	cout << endl;
	cout << "accumulate_operator: "; accumulate_operator.show();
	cout << endl;
	cout << "defuzzy_operator: "; defuzzy_operator.show();
	cout << endl;
	cout << "gamma   : " << gamma << endl;
	cout << "accuracy: " << accuracy << endl; };

int OperatorSet::write(ofstream *output) {
	*output << endl;
	importance_operator.write(output);
	and_operator.write(output);
	or_operator.write(output);
	certainty_operator.write(output);
	inference_operator.write(output);
	accumulate_operator.write(output);
	defuzzy_operator.write(output);
	*output << gamma << space;
	*output << accuracy << endl << endl;
	return 1; };

int OperatorSet::read(ifstream *input) {
	importance_operator.read(input);
	and_operator.read(input);
	or_operator.read(input);
	certainty_operator.read(input);
	inference_operator.read(input);
	accumulate_operator.read(input);
	defuzzy_operator.read(input);
	*input >> gamma;
	*input >> accuracy;
	return 1; };


