/************************************************************************
 * This library is free software; you can redistribute it and/or	*
 * modify it under the terms of the GNU Library General Public		*
 * License as published by the Free Software Foundation; either		*
 * version 2 of the License, or (at your option) any later version.	*
 * 									*
 * This library is distributed in the hope that it will be useful,	*
 * but WITHOUT ANY WARRANTY; without even the implied warranty of	*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU	*
 * Library General Public License for more details.			*
 * 									*
 * You should have received a copy of the GNU Library General Public	*
 * License along with this library; if not, write to the		*
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,		*
 * Boston, MA  02111-1307, USA.						*
 * 									*
 * The GNU Library General Public License file is included in		*
 * Online-Documentation							*
 *									*
 * GNU's Homepage - http://www.gnu.ai.mit.edu/home.html			*
 *									*
 * Questions & comments referring to *FLIP++: please contact us:	*
 *									*
 * Snail-Mail:								*
 *	Dr. Wolfgang SLANY						*
 *	Database and Expert System Group (E184-2)     			*
 *	Institute of Information Systems               			*
 *	Vienna University of Technology					*
 *	Paniglgasse 16, A-1040 Vienna, AUSTRIA				*
 * Telephone:								*
 *	+43-1-58801-6123     (area-code 1 maybe changed to 201 in 1998) *
 * FAX:									*
 *	+43-1-5055304	     (area-code 1 maybe changed to 201 in 1998) *
 * eMail:								*
 *	sflip@dbai.tuwien.ac.at						*
 * WWW:									*
 *	http://www.dbai.tuwien.ac.at/proj/StarFLIP/			*
 *                                                                      *
 * $Id: flipoper.h,v 1.4 1997/06/24 17:30:04 reichel Exp $ *
 ***********************************************************************/


//
// $Id: flipoper.h,v 1.4 1997/06/24 17:30:04 reichel Exp $
// 
//
//     ---------------------
//	F L I P O P E R . H
//     ---------------------
//

#ifndef flipoper
#define flipoper

#include "flipcvar.h"

enum fuzzy_operator {minimum,maximum,fuzzy_and,fuzzy_or,
		     ass_comp_and,ass_comp_or, product, sum,
		     no_fop};			 // to be continued
enum defuzzification_operator {cog,fimax,mom,no_dfop}; // to be continued
				// center of gravity
				// first maximum
				// mean of maximum
				// no defuzzy operator
enum concat_operator {c_and,c_or,AND,OR,no_cop}; // to select fuzzy_operator

const double pi=3.14159265359;

//
//
//	* * * * * * * * * *   F u z z y O p e r a t o r   * * *
//
//

class FuzzyOperator {

public:

// * * constructors:
	FuzzyOperator(void) { oper=no_fop; };
	FuzzyOperator(fuzzy_operator op) { oper=op; };

// * * complex methods:
	void show(void);
	int  write(ofstream *output);
	int  read(ifstream *input);

// * * simple methods to access data elements:
	void set_oper(fuzzy_operator new_op) { oper=new_op; };
	fuzzy_operator *get_oper(void)       { return &oper; };

protected:
// * * data elements:
	fuzzy_operator oper; };

//
//
//	* * * * * * * * * *   D e F u z z y O p e r a t o r   * * *
//
//

class DeFuzzyOperator {

public:

// * * constructors:
	DeFuzzyOperator(void) 			     { oper=no_dfop; };
	DeFuzzyOperator(defuzzification_operator op) { oper=op; };

// * * complex methods:
	void show(void);
	int  write(ofstream *output);
	int  read(ifstream *input);

// * * simple methods to access data elements:
	void set_oper(defuzzification_operator new_op) {oper=new_op;};
	defuzzification_operator *get_oper(void)       {return &oper;};

protected:
// * * data elements:
	defuzzification_operator oper; };

//
//
//	* * * * * * * * * *   C o n c a t O p e r a t o r   * * *
//
//

class ConcatOperator {

public:

// * * constructors:
	ConcatOperator(void) 		   { oper=no_cop; };
	ConcatOperator(concat_operator op) { oper=op; };

// * * complex methods:
	void show(void);
	int  write(ofstream *output);
	int  read(ifstream *input);
	int  read(strstream *input);
	int  compare(ConcatOperator *c1, ConcatOperator *c2);

// * * simple methods to access data elements:
	void set_oper(concat_operator new_op) { oper=new_op; };
	concat_operator *get_oper(void)       { return &oper; };

protected:
// * * data elements:
	concat_operator oper; };

//
//
//	* * * * * * * * * *   t - n o r m   * * *
//
//

class TNorm {

public:
	float min(FList *mvs);
	float min(float oamv,  float cer);
	void  min(float aocmv, PlimFuzzySet *fs); 

	float prod(FList *mvs);
	float prod(float oamv,  float cer);
	void  prod(float aocmv, PlimFuzzySet *fs); };

//
//
//	* * * * * * * * * *   t - c o - n o r m   * * *
//
//

class TCoNorm {

public:
	float max(FList *mvs);
	float max(float oamv,  float cer);
	void  max(float aocmv, PlimFuzzySet *fs); 

	float addition(FList *mvs);
	float addition(float oamv,  float cer);
	void  addition(float aocmv, PlimFuzzySet *fs); };

//
//
//	* * * * *    c o m p a n s a t o r y   -   o p e r a t o r    * * *
//
//

class CompOp {

public:
	float ass_comp_and(FList *mvs);
	float ass_comp_and(float oamv,  float cer);
	void  ass_comp_and(float aocmv, PlimFuzzySet *fs);

	float ass_comp_or(FList *mvs);
	float ass_comp_or(float oamv,  float cer);
	void  ass_comp_or(float aocmv, PlimFuzzySet *fs);

	float fuzzy_and(FList *mvs, float gamma);
	float fuzzy_and(float oamv,  float cer, float gamma);
	void  fuzzy_and(float aocmv, PlimFuzzySet *fs, float gamma);

	float fuzzy_or(FList *mvs, float gamma);
	float fuzzy_or(float oamv,  float cer, float gamma);
	void  fuzzy_or(float aocmv, PlimFuzzySet *fs, float gamma); };

//
//
//	* * * * * * * * * *   i n t e r s e c t i o n   * * *
//
//

//
//
//	* * * * * * * * * *   d e f u z z y   * * *
//
//

class DefOp {

public:
	float cog(CombinedVariable *cv);	// center of gravity
	float first_max(CombinedVariable *cv);	// maximum
	float mom(CombinedVariable *cv);	// mean of maximums
	void  show_cover(PList cover, int cover_number);
	float integrate(PList *cpl); };

//
//
//	* * * * * * * * * *   O p e r a t o r S e t  * * *
//
//
class   OperatorSet {
friend  float aggregate(float mvs[], concat_operator agg_op,
			OperatorSet *operator_set);
friend 	float certify(float oamv, float certainty,
			OperatorSet *operator_set);
friend 	void  inference(float oacmv, PlimFuzzySet *fs,
			OperatorSet *operator_set);
friend 	float accumulate(CombinedVariable *cv,
			OperatorSet *operator_set);


public:
// * * constructor:
	OperatorSet(void)
		{ importance_operator.set_oper(no_fop);
		  and_operator.set_oper(minimum);
		  or_operator.set_oper(maximum);
		  certainty_operator.set_oper(minimum);
		  inference_operator.set_oper(minimum);
		  accumulate_operator.set_oper(minimum);
		  defuzzy_operator.set_oper(cog);
		  gamma=1;
		  accuracy=10; };

// * * complex methods:
	void show(void);
	int  write(ofstream *output);
	int  read(ifstream *input);

// simple methods to access data elements:
	void set_importance_operator(fuzzy_operator op)
		{ importance_operator.set_oper(op); };
	void set_and_operator(fuzzy_operator op)
		{ and_operator.set_oper(op); };
	void set_or_operator(fuzzy_operator op)
		{ or_operator.set_oper(op); };
	void set_certainty_operator(fuzzy_operator op)
		{ certainty_operator.set_oper(op); };
	void set_inference_operator(fuzzy_operator op)
		{ inference_operator.set_oper(op); };
	void set_accumulate_operator(fuzzy_operator op)
		{ accumulate_operator.set_oper(op); };
	void set_defuzzy_operator(defuzzification_operator op)
		{ defuzzy_operator.set_oper(op); };
	void set_gamma(float g)     { gamma=g; };
	void set_accuracy(int ac) { accuracy=ac; };

	fuzzy_operator *get_importance_operator(void)
		{ return importance_operator.get_oper(); };
	fuzzy_operator *get_and_operator(void)
		{ return and_operator.get_oper(); };
	fuzzy_operator *get_or_operator(void)
		{ return or_operator.get_oper(); };
	fuzzy_operator *get_certainty_operator(void)
		{ return certainty_operator.get_oper(); };
	fuzzy_operator *get_inference_operator(void)
		{ return inference_operator.get_oper(); };
	fuzzy_operator *get_accumulate_operator(void)
		{ return accumulate_operator.get_oper(); };
	defuzzification_operator *get_defuzzy_operator(void)
		{ return defuzzy_operator.get_oper(); };
	float get_gamma(void)    { return gamma; };
	int   get_accuracy(void) { return accuracy; };

private:
	FuzzyOperator   importance_operator;
	FuzzyOperator   and_operator;
	FuzzyOperator   or_operator;
	FuzzyOperator   certainty_operator;
	FuzzyOperator   inference_operator;
	FuzzyOperator   accumulate_operator;
	DeFuzzyOperator defuzzy_operator;
	float		gamma;
	int             accuracy; };

//
//
//	* * * * * * * * * *   i n t e r f a c e    * * *
//
//

float join_mv_with_importance(float mv, float imp,OperatorSet *operator_set);
float aggregate(FList *mvs,concat_operator *agg_op,OperatorSet *operator_set);
float certify(float oamv, float certainty, OperatorSet *operator_set);
void  inference(float oacmv, PlimFuzzySet *fs, OperatorSet *operator_set);
float accumulate(CombinedVariable *cv, OperatorSet *operator_set);
int   calculate(CombinedVariable *cvr, CombinedVariable *cva, 
		CombinedVariable *cvb, concat_operator *agg_op,
		OperatorSet *operator_set);

#endif

