/************************************************************************
 * This library is free software; you can redistribute it and/or	*
 * modify it under the terms of the GNU Library General Public		*
 * License as published by the Free Software Foundation; either		*
 * version 2 of the License, or (at your option) any later version.	*
 * 									*
 * This library is distributed in the hope that it will be useful,	*
 * but WITHOUT ANY WARRANTY; without even the implied warranty of	*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU	*
 * Library General Public License for more details.			*
 * 									*
 * You should have received a copy of the GNU Library General Public	*
 * License along with this library; if not, write to the		*
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,		*
 * Boston, MA  02111-1307, USA.						*
 * 									*
 * The GNU Library General Public License file is included in		*
 * Online-Documentation							*
 *									*
 * GNU's Homepage - http://www.gnu.ai.mit.edu/home.html			*
 *									*
 * Questions & comments referring to *FLIP++: please contact us:	*
 *									*
 * Snail-Mail:								*
 *	Dr. Wolfgang SLANY						*
 *	Database and Expert System Group (E184-2)     			*
 *	Institute of Information Systems               			*
 *	Vienna University of Technology					*
 *	Paniglgasse 16, A-1040 Vienna, AUSTRIA				*
 * Telephone:								*
 *	+43-1-58801-6123     (area-code 1 maybe changed to 201 in 1998) *
 * FAX:									*
 *	+43-1-5055304	     (area-code 1 maybe changed to 201 in 1998) *
 * eMail:								*
 *	sflip@dbai.tuwien.ac.at						*
 * WWW:									*
 *	http://www.dbai.tuwien.ac.at/proj/StarFLIP/			*
 *                                                                      *
 * $Id: fliprule.cpp,v 1.4 1997/06/24 17:30:08 reichel Exp $ *
 ***********************************************************************/


//
// $Id: fliprule.cpp,v 1.4 1997/06/24 17:30:08 reichel Exp $
// 
//
//     ------------------------------------
//	F L I P R U L E . C C  ( . C P P )
//     ------------------------------------
//

#include "fliprule.h"
//
//
//	* * * * * * * * * *   T e r m  * * *
//
//

// * * complex methods:

void Term::show(void) {
	cout << cv->get_cv_name() << " IS " << lt << space; };

int Term::write(ofstream *output) {
	*output << cv->get_cv_name() << " IS " << lt << space;
	return 1; };

int Term::read(ifstream *input,CVList *cv_list) {
	name instr;
	*input >> instr; // cv
	cv=cv_list->find_cv(instr);
	// check if cv with this name exists
	if (cv == NULL) {
		cerr << "CombinedVariable in Rule not in cv_list!" << endl;
		return 2; };
	*input >> instr; // IS
	// read lt
	*input >> lt;
	return 1; };

void Term::copy(Term *t) {
	if (t != NULL) { cv=t->get_cv(); strcpy(lt,*(t->get_lt())); }; };

int Term::compare(Term *t1, Term *t2) {
	// returns 1 if terms are equal
	// else    0
	if (t1 == NULL || t2 == NULL) return 0;
	if (strcmp(*(t1->get_lt()),*(t2->get_lt())) == 0 && 	
	  strcmp(t1->get_cv()->get_cv_name(),t2->get_cv()->get_cv_name())==0) 
		return 1; 
	else	return 0; };

void Term::make_ShortString(strstream *ss) {
	// <tp> := <cvp> <ltp> ;
	// <cvp> := "SA" | "S_SOLL" ;
	// <ltp> := "very_low " | "low " | "medium " | "high " | "very_high " ;
	// cout << "char *Term::make_ShortString(void)" << endl;

	*ss << "1.";
	if 	(strcmp(cv->get_cv_name(),"SA") == 0)	*ss << "1.";
	else						*ss << "2.";
	if 	(strcmp(lt,"very_low") == 0)	*ss << "1.";
	else if (strcmp(lt,"low") == 0)		*ss << "2.";
	else if (strcmp(lt,"medium") == 0)	*ss << "3.";
	else if (strcmp(lt,"high") == 0)	*ss << "4.";
	else					*ss << "5.";
	// show();
	// cout << "T: " << ss->str() << endl;
	};

//
//
//	* * * * * * * * * *  T L i s t  * * *
//
//

// * * destructor:

TList::~TList(void) {
	while (first != NULL) {
		Term *t=first;
		first = t->get_next();
		delete t; };
	last=NULL;
	number=0; };

// * * complex methods:

void TList::show(void) {
	Term *element=first;
	while (element != NULL) {
		element->show();
		element=element->get_next(); }; };

void TList::show(ConcatOperator *op) {
	Term *element=first;
	if (element != NULL) {
		while (element->get_next() != NULL) {
			element->show();
			op->show();
			element=element->get_next(); };
		element->show(); }; };

int TList::write(ofstream *output, ConcatOperator *op) {
	Term *element=first;
	*output << "/* #: " << number << " */ ";
	if (element != NULL) {
		while (element->get_next() != NULL) {
			element->write(output);
			op->write(output);
			element=element->get_next(); };
		element->write(output); };
	return 1; };

int TList::read(ifstream *input, ConcatOperator *op, CVList *cv_list) {
	char instr[10];
	int n;
	*input >> instr; //  /*
	*input >> instr; //  #:
	*input >> n;
	*input >> instr; //  */
	for (int i=0; i<n; i++) {
		add();
		if (last->read(input,cv_list)==2) return 2;
		if (i<n-1) op->read(input); };
	return 1; };

void TList::add(CombinedVariable *ncv, name nlt) {
	Term *element=new Term;
	element->set_cv(ncv);
	element->set_lt(nlt);
	if (first == NULL ) { first=element;      last=element; }
	else 	       { last->set_next(element); last=element; };
	number++; };

void TList::add(Term *t) {
	if (first == NULL ) { first=t;      last=t; }
	else 	       { last->set_next(t); last=t; };
	number++; };

Term *TList::find(int index) {
	Term *element=first;
	if (index < number) {
		for (int i=0; i<index; i++)
			element=element->get_next();
	  	return element; };
	for (int j=number; j<=index; j++) add();
	return last; };

void TList::copy(TList *tl) {
	remove_all();
	if (tl != NULL) {
		Term *element=tl->get_first();
		for (int i=0; i<tl->get_number(); i++) {
			add();
			find(i)->copy(element);
			element=element->get_next(); }; }; };

int TList::compare(TList *tl1, TList *tl2) {
	// returns 1 if term lists are equal
	// else    0	
	if (tl1 == NULL || tl2 == NULL) return 0;
	int n1=tl1->get_number();
	int n2=tl2->get_number();
	if (n1 != n2) return 0;
	Term *element1=tl1->get_first();
	Term *element2=tl2->get_first();
	for (int i=0; i<n1; i++) {
		if (element1->compare(element1,element2) != 1) 
			return 0; 
		element1=element1->get_next();
		element2=element2->get_next(); }; 
	return 1; };

int TList::remove(int t_index) {
	if (t_index >= number) {
		cerr << "no such clause int TList::remove(int t_index)"
		     << endl;
		return -1; };
	Term *element;
	Term *to_delete;
	if (t_index==0) { to_delete=first;
			  first=first->get_next(); }
	else { element=find(t_index-1);
	       to_delete=element->get_next();
	       element->set_next(element->get_next()->get_next()); };
	delete to_delete;
	number--;
	return 1; };

void TList::remove_all(void) {
	while (first != NULL) {
		Term *element=first;
		first = element->get_next();
		delete element; };
	first=NULL;
	last=NULL;
	number=0; };

void TList::make_ShortString(strstream *ss) {
	// <tlp> := <tp> | <tlp> <tp> ;
	// cout << "char *TList::make_ShortString(void)" << endl;
	Term *element=first;

	for (int i=0; i<number-1; i++) *ss << "2.";
	*ss << "1.";
	while (element != NULL) {
		element->make_ShortString(ss);
		element=element->get_next(); };
	// show();
	// cout << "TL: " << ss->str() << endl;
	};

//
//
//	* * * * * * * * * *   R u l e   * * *
//
//

// * * destructor:

Rule::~Rule(void) {
	delete rs_term;
	delete ls_t_list;
	aggregate_operator=no_cop;
	certainty = -1;
	next=NULL; };

// * * complex methods:

void Rule::show(void) {
	if (rs_term->get_cv() != NULL) {
		cout << "IF ";
		if (ls_t_list!=NULL) ls_t_list->show(&aggregate_operator);
		cout << "THEN ";
		if (rs_term!=NULL) rs_term->show();
		cout << "WITH CERTAINTY " << certainty
		     << '.' << endl; }; };

int Rule::write(ofstream *output) {
	if (rs_term->get_cv() != NULL) {
		*output << "IF ";
		ls_t_list->write(output, &aggregate_operator);
		*output << "THEN ";
		rs_term->write(output);
		*output << "WITH CERTAINTY " << certainty
			<< '.' << endl; };
	return 1; };

int Rule::read(ifstream *input, CVList *cv_list) {
	name instr;
		// cout << "bin in Rule::read : ";
	*input >> instr; // IF
		// cout << instr << endl;
	if (ls_t_list->read(input,&aggregate_operator,cv_list)==2) return 2;
	*input >> instr; // THEN
	if (rs_term->read(input,cv_list)==2) return 2;
	*input >> instr; // WITH
	*input >> instr; // CERTAINTY
	*input >> certainty;
	if (certainty!=0 && certainty!=1) *input >> instr; // .
	// cout << "bin am normalen ende von Rule::read" << infloat << endl;
	return 1; };

void Rule::copy(Rule *r) {
	if (r != NULL) {
		set_rs_term(r->get_rs_term());
		set_ls_t_list(r->get_ls_t_list());
		aggregate_operator=*(r->get_aggregate_operator());
		certainty=r->get_certainty(); }; };

int Rule::compare(Rule *r1, Rule *r2) {
	// returns 1 if rules are equal
	// else    0	
	if (r1 == NULL || r2 == NULL) return 0;
	Term 		*t1  = r1->get_rs_term();
	Term 		*t2  = r2->get_rs_term();
	TList 		*tl1 = r1->get_ls_t_list();
	TList 		*tl2 = r2->get_ls_t_list();
	ConcatOperator 	*o1  = r1->get_aggregate_operator();
	ConcatOperator 	*o2  = r2->get_aggregate_operator();
	float		 c1  = r1->get_certainty();
	float		 c2  = r2->get_certainty();
	if ( 	t1->compare(t1,t2) 	== 1 &&
		tl1->compare(tl1,tl2) 	== 1 &&
		o1->compare(o1,o2) 	== 1 &&
		c1 == c2 ) 
			return 1; 
	return 0; };

void Rule::make_ShortString(strstream *ss) {
	// <br> := <tp> <ltc> <c> ;
	// <cr> := <o> <tlp> <ltc> <c> ;
	// <ltc> := "very_low " | "low " | "medium " | "high " | "very_high " ;
	// <o> := "AND " ;
	// <c> := "0.5 " | "1.0 " ;
	// cout << "char *Rule::make_ShortString(void)" << endl;
	name ltc = *(rs_term->get_lt());

	if (get_ls_t_list()->get_number() < 2) {
		// <br>
		*ss << "1.";	// <br>
		*ss << "1.";	// in <br>	
		ls_t_list->get_first()->make_ShortString(ss); }		
	else {
		// <cr>
		*ss << "2.";	// <cr>
		*ss << "1.";	// in <cr>
		*ss << "1.";	// <o>
		ls_t_list->make_ShortString(ss); };
	if 	(strcmp(ltc,"very_low") == 0)	*ss << "1.";
	else if (strcmp(ltc,"low") == 0)	*ss << "2.";
	else if (strcmp(ltc,"medium") == 0)	*ss << "3.";
	else if (strcmp(ltc,"high") == 0)	*ss << "4.";
	else					*ss << "5.";
	if (certainty <= 0.5)	*ss << "1.";
	else			*ss << "2."; 
	// show();
	// cout << "R: " << ss->str() << endl;
	}; 
	
// * * simple methods to access data elements:
void Rule::set_rs_term(Term *t) {
	if (rs_term==NULL) { Term *nt=new Term; rs_term=nt; };
	rs_term->copy(t); };

void Rule::set_ls_t_list(TList *tl) {
	if (ls_t_list==NULL) { TList *ntl=new TList; ls_t_list=ntl; };
	ls_t_list->copy(tl); };

//
//
//	* * * * * * * * * *   R u l e B l o c k    * * *
//
//

// * * destructor:

RuleBlock::~RuleBlock(void) {
	while (first != NULL) {
		Rule *r=first;
		first = r->get_next();
		delete r; };
	last=NULL;
	number=0; };

// * * complex methods:

void RuleBlock::show(void) {
	Rule *element=first;
	while (element != NULL) {
		element->show();
		element=element->get_next(); }; };

int RuleBlock::write(ofstream *output) {
	Rule *element=first;
	*output << "/* #: " << number << " */" << endl;
	while (element != NULL) {
		element->write(output);
		element=element->get_next(); };
	return 1; };

int RuleBlock::read(ifstream *input, CVList *cv_list) {
	char instr[10];
	int n;
	*input >> instr; //  /*
	*input >> instr; //  #:
	*input >> n;
	*input >> instr; //  */
	for (int i=0; i<n; i++) {
		add();
		if (last->read(input,cv_list)==2) return 2; };
	return 1; };

void RuleBlock::add(Rule *r) {
	Rule *element=new Rule;
	element->copy(r);
	if (find(r) < 0) {
		if (first == NULL ) { first=element;           last=element; }
		else 	            { last->set_next(element); last=element; };
		number++; }; };

Rule *RuleBlock::find(int index) {
	Rule *element=first;
	if (index < number) {
		for (int i=0; i<index; i++)
			element=element->get_next();
		return element; };
	// for (int j=number; j<=index; j++) add();
	return NULL; };

int  RuleBlock::find(Rule *r) {
	// returns index of rule or -1
	if (r == NULL) return -1;
	Rule *element=get_first();
	for (int i=0; i<get_number(); i++) {
		if (r->compare(r,element) == 1) 
			return i; 
		element=element->get_next(); };
	return -1; };	

void RuleBlock::copy(RuleBlock *rb) {
	remove_all();
	if (rb != NULL) {
		Rule *element=rb->get_first();
		for (int i=0; i<rb->get_number(); i++) {
			add();
			find(i)->copy(element);
			element=element->get_next(); }; }; };

int RuleBlock::remove(int r_index) {
	if (r_index >= number) {
		cerr << "no such Rule in RuleBlock::remove(int r_index)" 
		     << endl;
		return -1; };
	Rule *element;
	Rule *to_delete;
	if (r_index==0) { to_delete=first;
			  first=first->get_next(); }
	else { element=find(r_index-1);
	       to_delete=element->get_next();
	       element->set_next(element->get_next()->get_next()); };
	delete to_delete;
	number--;
	return 1; };

void RuleBlock::remove_all(void) {
	while (first != NULL) {
		Rule *element=first;
		first = element->get_next();
		delete element; };
	first=NULL;
	last=NULL;
	number=0; };

void RuleBlock::make_ShortString (strstream *ss) {
	// <rb> := <cvc> <rl>;
	// <rl> := <r> | <rl> <r> ;
	// <cvc> := "LOG" ;
	//cout << "char *RuleBlock::make_ShortString(void)" << endl;
	Rule *element=first;

	*ss << "1.";
	for (int i=0; i<number-1; i++) *ss << "2.";
	*ss << "1.";
	while (element != NULL) {
		element->make_ShortString(ss);
		element=element->get_next(); };
	//show();
	//cout << "RB: " << ss->str() << endl;
	};

//
//
//	* * * * * * * * * *   R u l e S e t   * * *
//
//

// * * destructor:

RuleSet::~RuleSet(void) {
	while (first != NULL) {
		RuleBlock *rb=first;
		first = rb->get_next();
		delete rb; };
	delete ps_list; 
	delete cv_list;
	delete operator_set;  };

// * * complex methods:

void RuleSet::show(void) {
	RuleBlock *element=first;
	cout << rs_name << endl;
	while (element != NULL) {
		element->show();
		element=element->get_next(); }; };

int RuleSet::write(ofstream *output) {
	ps_list->write(output);
	cv_list->write(output);
	operator_set->write(output);
	RuleBlock *element=first;
	*output << rs_name << endl;
	*output << "/* #: " << number << " */" << endl;
	while (element != NULL) {
		element->write(output);
		element=element->get_next(); };
	return 1; };

int RuleSet::read(ifstream *input) {
	ps_list->read(input);
	cv_list->read(input);
	operator_set->read(input);
	*input >> rs_name;
	char instr[40];
	int n;
	*input >> instr; //  /*
	*input >> instr; //  #:
	*input >> n;
	*input >> instr; //  */
	for (int i=0; i<n; i++) {
		add();
		if (last->read(input,cv_list)==2) return 2; };
	return 1; };

void RuleSet::add(RuleBlock *rb) {
	RuleBlock *element=new RuleBlock;
	if (strcmp(rs_name,"empty_rule_set")==0)
		strcpy(rs_name,"nameless_rule_set");
	element->copy(rb);
	if (first == NULL ) { first=element;      last=element; }
	else 	       { last->set_next(element); last=element; };
	number++; };

void RuleSet::add_all_RuleBlocks(RuleSet *rs) {
	RuleBlock *element=rs->get_first();
	while (element != NULL) {
		add(element);
		element=element->get_next(); }; };

int RuleSet::add_Rule(TList *ls_tl,Term *rs_t,concat_operator op,float cer) {
	Rule *r=new Rule;
	concat_operator oper;
	CombinedVariable *cv;
	int n,index;

	if (strcmp(rs_name,"empty_rule_set")==0)
		strcpy(rs_name,"nameless_rule_set");
	if      (op==c_and || op==AND) oper=AND;
	else if (op==c_or  || op==OR)  oper=OR;
	else    oper=no_cop;
	r->set_ls_t_list(ls_tl);
	r->set_rs_term(rs_t);
	r->set_aggregate_operator(oper);
	r->set_certainty(cer);

	// find fitting RuleBlock or next empty ruleblock
	int rb_index = search(rs_t->get_cv());
	if ( rb_index < 0 ) {
		rb_index = number;
		add(); };
	find(rb_index)->add(r);
	delete r;
	// cvs in cv_list:
	n=ls_tl->get_number();
	for (int i=0; i<n; i++) {
		cv=ls_tl->find_cv(i);
		index=cv_list->find_cv_by_name(cv);
		if (index < 0) cv_list->add(cv); };
	cv=rs_t->get_cv();
	index=cv_list->find_cv_by_name(cv);
	if (index < 0) cv_list->add(cv);
	return 1; };

int RuleSet::add_BasicRule(CombinedVariable *left_cv, name left_lt,
		  CombinedVariable *right_cv, name right_lt, float cer) {
	Term  *tr  = new Term;
	TList *tli = new TList;
	tli->add(left_cv,left_lt);
	tr->set_cv(right_cv);
	tr->set_lt(right_lt);
	add_Rule(tli,tr,no_cop,cer);
	delete tr;
	delete tli;
	return 1; };

int RuleSet::add_ComplexRule(CombinedVariable *left_cv_1, name left_lt_1,
			     CombinedVariable *left_cv_2, name left_lt_2,
			     CombinedVariable *right_cv,  name right_lt,
			     concat_operator  agg_op,     float cer) {
	Term  *tr  = new Term;
	TList *tli = new TList;
	tli->add(left_cv_1,left_lt_1);
	tli->add(left_cv_2,left_lt_2);
	tr->set_cv(right_cv);
	tr->set_lt(right_lt);
	add_Rule(tli,tr,agg_op,cer);
	delete tr;
	delete tli;
	return 1; };

RuleBlock *RuleSet::find(int index) {
	RuleBlock *element=first;
	if (index < number) {
		for (int i=0; i<index; i++)
			element=element->get_next();
		return element; }
	else 	return NULL; };

int RuleSet::search(CombinedVariable *cv) {
	// find cv on right side of a ruleblock in the ruleset
	// and return its index
	// key for search is the name of the cv, and not its address
	int i=0;
	RuleBlock *element=first;
	while (element != NULL) {
		if (element->get_number() < 1) return -1;
		if (strcmp( 	
			element->get_right_side_cv()->get_cv_name(),
			cv->get_cv_name())==0) 		return i;
		element=element->get_next();
		i++; };
	return -1; };

CombinedVariable *RuleSet::get_cv(name cv_name) {
	// - when add rule, write the concerned cvs
	//   into cv_list (or other CVList)
	// - then just search in cv_list
	return cv_list->find_cv(cv_name); };

int RuleSet::remove(int rb_index) {
	if (rb_index >= number) {
		cerr << "no such RuleBlock in RuleSet::remove(int rb_index)" 
		     << endl;
		return -1; };
	RuleBlock *element;
	RuleBlock *to_delete;
	if (rb_index==0) { to_delete=first;
			   first=first->get_next(); }
	else { element=find(rb_index-1);
	       to_delete=element->get_next();
	       element->set_next(element->get_next()->get_next()); };
	delete to_delete;
	number--;
	return 1; };

int RuleSet::remove(CombinedVariable *which) {
	// find fitting RuleBlock
	int rb_index = search(which);
	if (rb_index < 0) {
		return -1; };
	return remove(rb_index); }

void RuleSet::remove_cv_and_its_approaches(CombinedVariable *cv) {
	// removes cv
	// removes ruleblock with it on the right side
	// searches for all rules that include it
	// 	if it is only clause on left side (simple rule)
	//		then remove this rule
	//	else	remove this clause in rule
	char      *cv_name = cv->get_cv_name();
	int       cv_index = cv_list->find_cv_by_name(cv);
	RuleBlock *rb=first,*rbn;
	Rule      *r,*rn;
	Term      *t;
	int       rbi=0,ri,ti;

	while (rb != NULL) {
	    rbn = rb->get_next();
	    r  = rb->get_first();
	    ri = 0;
	    while (r != NULL) {
	    	rn = r->get_next(); 
		if (r->get_ls_t_list_number() < 2) {
		   if (strcmp(r->get_left_side_cv(0)->get_cv_name(),
		   	      cv_name)==0) 
		       	rb->remove(ri); }
		else {
		   t  = r->get_ls_t_list()->get_first();
		   ti = 0;
		   while (t != NULL) {
			if (strcmp(t->get_cv()->get_cv_name(),cv_name)==0)
			   r->get_ls_t_list()->remove(ti);
			t = t->get_next();
			ti++; };
		    ri++; };
		r = rn; };
	    if (rb->get_number() < 1) 		remove(rbi);
	    else 				rbi++; 
	    rb = rbn; }; 

	remove(cv); 
	cv_list->remove(cv_index); };

void RuleSet::remove_all(void) {
	while (first != NULL) {
		RuleBlock *element=first;
		first = element->get_next();
		delete element; };
	first=NULL;
	last=NULL;
	number=0; };

float RuleSet::evaluate(CombinedVariable *cv) {
	float crisp_value;
	cv=get_cv(cv->get_cv_name());
	int ir = internal_evaluate(cv);
	if (ir > 0) {
		crisp_value = accumulate(cv, operator_set);
		cv->set_crisp_value(crisp_value); 
		return crisp_value; };
	return cv->get_crisp_value(); };

int RuleSet::internal_evaluate(CombinedVariable *cv) {
	int r_rb_index;       // right_RuleBlock_index
	int l_rb_index;       // left_RuleBlock_index
	int fs_index;	      // index of fuzzy_sets of linguistic term
	int op_number;	      // number of operands in one RuleBlock
	int r_number;	      // number rules in one RuleBlock
	int r_index;          // Rule index
	int op_index;         // Operand index
	int matched;          // 0 when cv allready matched else -1
	float fs_mv;	      // match_value from fs from cv
	float oamv,oacmv;     // over_all_(certivied)_match_value
	name lt;	      // linguistic_term
	FList *mvs=new FList; // match_values of left_side fuzzy_sets
	PlimFuzzySet *fs;     // fuzzy_set of right side to inference
	RuleBlock *rb;        // RuleBlock with cv on its right side

	if (cv==NULL) { cerr << "NULL pointer cannot be evaluated" << endl; 
			return -1; };
	cv->set_crisp_value(-1);
	r_rb_index = search(cv);     // RuleBlock_number according to cv
	if (r_rb_index < 0) { cerr << "cv not in RuleSet" << endl; 
			      return -1; };
	rb=find(r_rb_index);
	r_number=rb->get_number();
	for (r_index=0; r_index<r_number; r_index++) {
	  op_number=rb->get_operand_number(r_index);
	  for (op_index=0; op_index<op_number; op_index++) {
		// search for left side cv RuleBlock in RuleSet:
		CombinedVariable *locv,*lncv;
		locv=rb->get_left_side_cv(r_index,op_index);
		lncv=get_cv(locv->get_cv_name());
		l_rb_index=search(lncv);
		if (l_rb_index<0) { // no such RuleBlock in ruleset
			matched=lncv->check_match();
			if ( matched < 0 ) {
				cerr << "not instanciated cv" << endl;
				return -1; }; } // end if
		else	// there is a RuleBlock for left side cv
		     	internal_evaluate(lncv); 
	  }; }; // end for

	// set all match values to zero
	// for each rule of RuleBlock  a g g r e g a t e  all match_values
	//	of all operands to one over all match_value and
	//      certivy it with the certainty of each rule
	// 	then  i n f e r e n c e  this over all match_value with the
	// 	concerned fuzzy_set of the right_side CombinedVariable
	// ( a c c u m u l a t e  ( no action! ) all new matched     )
	// ( and instanciated                                        )
	// ( fuzzy_set of this right_side cv to one CombinedVariable )
	// ( 	this has not to happen, because the CombinedVariable )
	// (	consits of a number of fuzzy_sets )

	cv->set_all_mvs(0);
	for (r_index=0; r_index<r_number; r_index++) {
	  mvs->remove_all();
	  op_number=rb->get_operand_number(r_index);
	  if (op_number>1) {
	    for (op_index=0; op_index<op_number; op_index++) {
		// need index of fuzzy_set of linguistic_term
		strcpy(lt,*(rb->get_left_side_lt(r_index,op_index)));
		CombinedVariable *locv,*lncv;
		locv=rb->get_left_side_cv(r_index,op_index);
		lncv=get_cv(locv->get_cv_name());
		fs_index  = lncv->get_fs_index(lt);
		fs_mv     = lncv->get_fs_match_value(fs_index);
		float imp = lncv->get_importance();
		float mv  = join_mv_with_importance(fs_mv,imp,operator_set);
		mvs->add(mv); }; // end for op_index

	     oamv  = aggregate(mvs,rb->get_aggregate_operator(r_index),
			operator_set); } // end if
	     else {
		strcpy(lt,*(rb->get_left_side_lt(r_index,0)));
		CombinedVariable *locv,*lncv;
		locv=rb->get_left_side_cv(r_index,0);
		lncv=get_cv(locv->get_cv_name());
		fs_index  = lncv->get_fs_index(lt);
		oamv=lncv->get_fs_match_value(fs_index); }; 
	     // end else
	     oacmv = certify(oamv,rb->get_certainty(r_index),operator_set);
	     // need fuzzy_set of linguistic_term
		CombinedVariable *rocv,*rncv;
		rocv=rb->get_right_side_cv(r_index);
		rncv=get_cv(rocv->get_cv_name());	     
		strcpy(lt,*(rb->get_right_side_lt(r_index)));
		fs_index = rncv->get_fs_index(lt);
		fs       = rncv->find(fs_index);
		inference(oacmv, fs, operator_set); };
		delete mvs;
		return 1; };

void RuleSet::make_ShortString(strstream *ss) {
	// <rs> := <rb> | <rs> <rb> ;
	// cout << "char *RuleSet::make_ShortString(void)" << endl;
	
	// show();
	*ss << "1.";
	RuleBlock *element=first;
	// element->show();
	for (int i=0; i<number-1; i++) *ss << "2.";
	*ss << "1.";
	while (element != NULL) {
		element->make_ShortString(ss);
		element=element->get_next(); };
	*ss << '\0'; };


