/************************************************************************
 * This library is free software; you can redistribute it and/or	*
 * modify it under the terms of the GNU Library General Public		*
 * License as published by the Free Software Foundation; either		*
 * version 2 of the License, or (at your option) any later version.	*
 * 									*
 * This library is distributed in the hope that it will be useful,	*
 * but WITHOUT ANY WARRANTY; without even the implied warranty of	*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU	*
 * Library General Public License for more details.			*
 * 									*
 * You should have received a copy of the GNU Library General Public	*
 * License along with this library; if not, write to the		*
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,		*
 * Boston, MA  02111-1307, USA.						*
 * 									*
 * The GNU Library General Public License file is included in		*
 * Online-Documentation							*
 *									*
 * GNU's Homepage - http://www.gnu.ai.mit.edu/home.html			*
 *									*
 * Questions & comments referring to *FLIP++: please contact us:	*
 *									*
 * Snail-Mail:								*
 *	Dr. Wolfgang SLANY						*
 *	Database and Expert System Group (E184-2)     			*
 *	Institute of Information Systems               			*
 *	Vienna University of Technology					*
 *	Paniglgasse 16, A-1040 Vienna, AUSTRIA				*
 * Telephone:								*
 *	+43-1-58801-6123     (area-code 1 maybe changed to 201 in 1998) *
 * FAX:									*
 *	+43-1-5055304	     (area-code 1 maybe changed to 201 in 1998) *
 * eMail:								*
 *	sflip@dbai.tuwien.ac.at						*
 * WWW:									*
 *	http://www.dbai.tuwien.ac.at/proj/StarFLIP/			*
 *                                                                      *
 * $Id: fliprule.h,v 1.4 1997/06/24 17:30:09 reichel Exp $ *
 ***********************************************************************/


//
// $Id: fliprule.h,v 1.4 1997/06/24 17:30:09 reichel Exp $
// 
//
//     ---------------------
//	F L I P R U L E . H
//     ---------------------
//

#ifndef fliprule
#define fliprule

#include "flipoper.h"

//
//
//	* * * * * * * * * *   T e r m  * * *
//
//

class Term {
public:
// * * constructor:
	Term(void) {  cv = NULL;
		      strcpy(lt,"no_linguistic_term");
		      next = NULL; };
	~Term(void) { cv = NULL;
		      strcpy(lt," ");
		      next = NULL; };

// * * complex methods:
	void show(void);
	int  write(ofstream *output);
	int  read(ifstream *input, CVList *cv_list);
	void copy(Term *t);	
	int  compare(Term *t1, Term *t2);
	void make_ShortString(strstream *ss);

// * * simple methods to access data elements:
	void set_cv(CombinedVariable *ncv) { cv=ncv; };
	void set_lt(name nlt)              { strcpy(lt,nlt); };
	void set_next(Term *t)		   { next=t; };

	CombinedVariable *get_cv(void)     { return cv; };
	name 		 *get_lt(void)     { return &lt; };
	Term		 *get_next(void)   { return next; };

private:
// * * data elements:
	CombinedVariable  *cv;
	name 	          lt;
	Term		  *next; };

//
//
//	* * * * * * * * * *   T L i s t * * *
//
//

class TList {
public:
// * * constructor:
	TList(void) { first=NULL; last=NULL; number=0; };
	~TList(void);

// * * complex methods:
	void show(void);
	void show(ConcatOperator *op);
	int  write(ofstream *output, ConcatOperator *op);
	int  read(ifstream *input,   ConcatOperator *op, CVList *cv_list);
	void add(void) 			  { add(NULL,"no_linguistic_term"); };
	void add(Term *t); 
	void add(CombinedVariable *cv, name lt="no_linguistic_term");
	Term *find(int index);
	CombinedVariable *find_cv(int index) { return find(index)->get_cv();};
	name *find_lt(int index)             { return find(index)->get_lt();};
	Term *operator [] (int index)        { return find(index); };
	void copy(TList *tl);
	int  compare(TList *tl1, TList *tl2);
	int  remove(int t_index);
	void remove_all(void);
	void make_ShortString(strstream *ss);

// * * simple methods to access lower class data elements:
	void set_cv(CombinedVariable *ncv, int t_index)
		{ find(t_index)->set_cv(ncv); };
	void set_lt(name nlt, int t_index) { find(t_index)->set_lt(nlt); };

	CombinedVariable *get_cv(int t_index)
		{ return find(t_index)->get_cv(); };
	name *get_lt(int t_index) { return find(t_index)->get_lt(); };

// * * simple methods to access data elements
	Term *get_first(void)   { return first; };
	Term *get_last(void)    { return last; };
	int  get_number(void)   { return number; };

protected:
	Term *first;
	Term *last;
	int number; };

//
//
//	* * * * * * * * * *   R u l e  * * *
//
//
class Rule {
public:
// * * constructor:
	Rule(void)
	      { rs_term=new Term;
		ls_t_list=new TList;
		certainty=0;
		aggregate_operator.set_oper(no_cop);
		next=NULL;};
	~Rule(void);

// * * complex methods:
	void show(void);
	int  write(ofstream *output);
	int  read(ifstream *input, CVList *cv_list);
	void copy(Rule *r);
	int  compare(Rule *r1, Rule *r2);
	void make_ShortString(strstream *ss);

// * * simple methods to access lower class data elements:
	void set_left_side_cv(int op_index, CombinedVariable *cv)
		{ ls_t_list->find(op_index)->set_cv(cv); };
	void set_right_side_cv(CombinedVariable *cv) { rs_term->set_cv(cv); };
	void set_left_side_lt(int op_index, name lt)
		{ ls_t_list->find(op_index)->set_lt(lt); };
	void set_right_side_lt(name lt) { rs_term->set_lt(lt); };
	void set_aggregate_operator(concat_operator agg_op)
		{ aggregate_operator.set_oper(agg_op); };

	CombinedVariable *get_left_side_cv(int cv_number)
		{ return ls_t_list->find(cv_number)->get_cv(); };
	CombinedVariable *get_right_side_cv(void) {return rs_term->get_cv();};
	name  *get_left_side_lt(int cv_number)
		{ return ls_t_list->find(cv_number)->get_lt(); };
	name  *get_right_side_lt(void) { return rs_term->get_lt(); };
	concat_operator *get_aggregate_operator_op(void)
		{ return aggregate_operator.get_oper(); };
	int   get_ls_t_list_number(void)
		{ return ls_t_list->get_number(); };

// * * simple methods to access data elements:
	void set_rs_term(Term *t);
	void set_ls_t_list(TList *tl);
	void set_aggregate_operator(ConcatOperator *agg_op)
		{ aggregate_operator=*agg_op; };
	void set_certainty(float cer)	{ certainty=cer; };
	void set_next(Rule *r)          { next=r; };

	Term  *get_rs_term(void)        { return rs_term; };
	TList *get_ls_t_list(void)      { return ls_t_list; };
	float get_certainty(void)       { return certainty; };
	ConcatOperator *get_aggregate_operator(void)
		{ return &aggregate_operator; };
	Rule  *get_next(void)		{ return next; };

protected:
	Term           *rs_term;
	TList          *ls_t_list;
	float          certainty;
	ConcatOperator aggregate_operator; // to combine single clauses
	Rule *next; };

//
//
//	* * * * * * * * * *   R u l e B l o c k   * * *
//
//
class RuleBlock {
public:
// * * constructor:
	RuleBlock(void) { first=NULL; last=NULL; number=0; next=NULL; };
	~RuleBlock(void);

// * * complex methods:
	void show(void);
	int  write(ofstream *output);
	int  read(ifstream  *input, CVList *cv_list);
	void add(void) 		      {add(NULL);};
	void add(Rule *r);
	Rule *find(int index);
	int  find(Rule *r);
	CombinedVariable *find_cv(int index)
		{ return find(index)->get_rs_term()->get_cv(); };
	name *find_lt(int index)
		{ return find(index)->get_rs_term()->get_lt();};
	Rule *operator [] (int index) {return find(index);};
	void copy(RuleBlock *rb);
	int  remove(int r_index);
	void remove_all(void);
	void make_ShortString(strstream *ss);

// * * simple methods to access lower class data elements
	void set_left_side_cv(int r_index,int op_index,CombinedVariable *cv)
		{ find(r_index)->set_left_side_cv(op_index,cv); };
	void set_right_side_cv(int r_index,CombinedVariable *cv)
		{ find(r_index)->set_right_side_cv(cv); };
	void set_left_side_lt(int r_index,int op_index,name lt)
		{ find(r_index)->set_left_side_lt(op_index,lt); };
	void set_right_side_lt(int r_index,name lt)
		{ find(r_index)->set_right_side_lt(lt); };
	void set_aggregate_operator(int r_index,concat_operator agg_op)
		{ find(r_index)->set_aggregate_operator(agg_op); };
	void set_certainty(int r_index,float cer)
		{ find(r_index)->set_certainty(cer); };

	CombinedVariable *get_left_side_cv(int r_index,int cv_number)
		{ return find(r_index)->get_left_side_cv(cv_number); };
	CombinedVariable *get_right_side_cv(int r_index=0)
		{ return find(r_index)->get_right_side_cv(); };
	name  *get_left_side_lt(int r_index,int cv_number)
		{ return find(r_index)->get_left_side_lt(cv_number); };
	name  *get_right_side_lt(int r_index)
		{ return find(r_index)->get_right_side_lt(); };
	concat_operator *get_aggregate_operator(int r_index)
		{ return find(r_index)->get_aggregate_operator_op(); };
	float get_certainty(int r_index)
		{ return find(r_index)->get_certainty(); };
	int   get_operand_number(int r_index)
		{ return find(r_index)->get_ls_t_list_number(); };

// * * simple methods to access data elements
	void      set_first(Rule *t)      { first=t; };
	void      set_last(Rule *t)       { last=t; };
	void      set_number(int n)       { number=n; };
	void	  set_next(RuleBlock *rb) { next=rb; };

	Rule      *get_first(void)        { return first; };
	Rule      *get_last(void)         { return last; };
	int       get_number(void)        { return number; };
	RuleBlock *get_next(void)         { return next; };

protected:
	Rule      *first;
	Rule      *last;
	int       number;
	RuleBlock *next; };

//
//
//	* * * * * * * * * *   R u l e S e t   * * *
//
//
class RuleSet {

public:
// * * constructor:
	RuleSet(void) { ps_list = new PSList; 
			cv_list = new CVList;
			operator_set = new OperatorSet; 
			first=NULL; last=NULL; number=0;
			strcpy(rs_name,"empty_rule_set"); };

	RuleSet(PSList *psl, CVList *cvl, OperatorSet *os) {
		ps_list=psl; cv_list=cvl; operator_set=os;
		first=NULL; last=NULL; number=0;
		strcpy(rs_name,"empty_rule_set"); };
	~RuleSet(void);

// * * complex methods:
	void show(void);
	int  write(ofstream *output);
	int  read(ifstream  *input);
	void add(RuleBlock *rb=NULL);
	void add_all_RuleBlocks(RuleSet *rs);
	int  add_Rule(TList *ls_tl,Term *rs_t,concat_operator op, float cer);
	int  add_BasicRule(CombinedVariable *left, name left_ling,
		CombinedVariable *right, name right_ling, float cer=1);
	int  add_ComplexRule(CombinedVariable *left_cv_1, name left_lt_1,
			    CombinedVariable *left_cv_2, name left_lt_2,
			    CombinedVariable *right_cv, name right_lt,
			    concat_operator agg_op, float cer=1);
	RuleBlock *find(int index);
	CombinedVariable *find_cv(int index)
		{return find(index)->get_first()->get_rs_term()->get_cv();};
	RuleBlock *operator [] (int index) {return find(index);};
	int   search(CombinedVariable *cv);
	int   remove(int rb_index);
	int   remove(CombinedVariable *which);
	void  remove_cv_and_its_approaches(CombinedVariable *which);
	void  remove_all(void);
	CombinedVariable *get_cv(name cv_name);

	int   calculate(CombinedVariable *cva, CombinedVariable *cvb,
			concat_operator *agg_op);;
	float evaluate(CombinedVariable *cv);
	int   internal_evaluate(CombinedVariable *cv);
	void  make_ShortString(strstream *ss);

// * * simple methods to access lower class data elements:
	void set_left_side_cv(int rb_index, int r_index, int op_index,
			CombinedVariable *cv)
		{ find(rb_index)->set_left_side_cv(r_index,op_index,cv); };
	void set_left_side_lt(int rb_index,int r_index,int op_index,name lt)
		{ find(rb_index)->set_left_side_lt(r_index,op_index,lt); };
	void set_right_side_cv(int rb_index,int r_index,CombinedVariable *cv)
		{ find(rb_index)->set_right_side_cv(r_index,cv); };
	void set_right_side_lt(int rb_index, int r_index, name lt)
		{ find(rb_index)->set_right_side_lt(r_index,lt); };
	void set_aggregate_operator(int rb_index,int r_index,
			concat_operator agg_op)
		{ find(rb_index)->set_aggregate_operator(r_index,agg_op); };
	void set_certainty(int rb_index, int r_index, float cer)
		{ find(rb_index)->set_certainty(r_index,cer); };

	CombinedVariable *get_left_side_cv(int rb_index,int r_index,
			int op_index)
		{ return find(rb_index)->get_left_side_cv(r_index,op_index);};
	name *get_left_side_lt(int rb_index, int r_index, int op_index )
		{ return find(rb_index)->get_left_side_lt(r_index,op_index);};
	CombinedVariable *get_right_side_cv(int rb_index, int r_index=0 )
		{ return find(rb_index)->get_right_side_cv(r_index); };
	name *get_right_side_lt(int rb_index, int r_index)
		{ return find(rb_index)->get_right_side_lt(r_index); };
	concat_operator *get_aggregate_operator(int rb_index, int r_index)
		{ return find(rb_index)->get_aggregate_operator(r_index); };
	float get_certainty(int rb_index, int r_index)
		{ return find(rb_index)->get_certainty(r_index); };
	int get_operand_number(int rb_index, int r_index)
		{ return find(rb_index)->get_operand_number(r_index); };
	int get_rule_number(int rb_index)
		{ return find(rb_index)->get_number(); };
	int get_rule_block_number(void) { return number; };

// * * simple methods to access data elements
	void  set_rs_name(name nn)               { strcpy(rs_name,nn); };
	void  set_ps_list(PSList *npsl)          { ps_list=npsl; };
	void  set_cv_list(CVList *ncvl)          { cv_list=ncvl; };
	void  set_operator_set(OperatorSet *nos) { operator_set=nos; };
	void  set_first(RuleBlock *t) 	         { first=t; };
	void  set_last(RuleBlock *t)  	         { last=t; };
	void  set_number(int n)       	         { number=n; };

	name	    *get_rs_name(void)        { return &rs_name; };
	PSList      *get_ps_list(void)       { return ps_list; };
	CVList      *get_cv_list(void)       { return cv_list; };
	OperatorSet *get_operator_set(void)  { return operator_set; };
	RuleBlock   *get_first(void)         { return first; };
	RuleBlock   *get_last(void)          { return last; };
	int         get_number(void)         { return number; };

private:
// * * data elements:
	name        rs_name;
	PSList      *ps_list;
	CVList      *cv_list;
	OperatorSet *operator_set;
	RuleBlock   *first;
	RuleBlock   *last;
	int         number; };

# endif
